#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

r"""Utilities to obtain identifiers that are unique at different contexts.

Contexts could be global, host local or application local.  All standard
`uuid`:mod: tools are included in this one: `UUID`:class:, `uuid1`:func:,
`uuid3`:func:, `uuid4`:func:, `uuid5`:func:, `getnode`:func: and standard
UUIDs constants `NAMESPACE_DNS`, `NAMESPACE_URL`, `NAMESPACE_OID` and
`NAMESPACE_X500`.

This module also contains:

- `str_uuid`:func:\ : Return a string with a GUID representation, random if
  the argument is True, or a host ID if not.

.. versionadded:: 1.7.0

.. deprecated:: 2.1.0

"""

from uuid import (
    UUID,
    uuid1,
    uuid3,
    uuid4,
    uuid5,
    getnode,  # noqa
    NAMESPACE_DNS,
    NAMESPACE_URL,
    NAMESPACE_OID,
    NAMESPACE_X500,
)


def str_uuid(random=False):  # pragma: no cover
    """Return a "Global Unique ID" as a string.

    :param random: If True, a random uuid is generated (does not use host id).

    .. deprecated:: 2.1.0 Use `uuid.uuid4`:func: or `uuid.uuid1`:func:.

    """
    fn = uuid4 if random else uuid1
    return str(fn())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#


def test_nameof():
    from xotl.tools.tools import nameof

    class foobar:
        pass

    singletons = (None, True, False, Ellipsis, NotImplemented)

    assert nameof(foobar) == nameof(foobar()) == "foobar"
    assert nameof(object) == "object"
    assert nameof(test_nameof) == "test_nameof"

    assert nameof(lambda x: x) == "<lambda>"

    assert [nameof(s) for s in singletons] == [
        "None",
        "True",
        "False",
        "Ellipsis",
        "NotImplemented",
    ]

    assert nameof(1) == "int"
    assert nameof(1.0) == "float"


# TODO: Add tests for remainder functions in this module.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Integer encoding and decoding in different bases.

"""

_DEFAULT_TABLE = "0123456789" "abcdefghijklmnopqrstuvwxyz" "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

_MAX_BASE = len(_DEFAULT_TABLE)

_DEFAULT_BASE = _MAX_BASE


def _check_base(base):
    """Check a base to be used in string to integer conversions.

    Return a tuple (base, table) if valid or raise an exception.

    """
    if isinstance(base, int):
        table = _DEFAULT_TABLE
        if not (1 < base <= _MAX_BASE):
            raise ValueError("`base` must be between 2 and %s" % _MAX_BASE)
    elif isinstance(base, str):
        table = base
        base = len(table)
    else:
        msg = (
            "`base` must be an integer (base) or a string (table) with "
            'length greater or equal to 2; %s "%s" given'
        )
        raise TypeError(msg % (type(base).__name__, base))
    return base, table


def int2str(number, base=_DEFAULT_BASE):
    """Return the string representation of an integer using a base.

    :param base: The base.
    :type base: Either an integer or a string with a custom table.

    Examples::

      >>> int2str(65535, 16)
      'ffff'

      >>> int2str(65535)
      'h31'

      >>> int2str(65110208921, 'merchise')
      'ehimseiemsce'

      >>> int2str(651102, 2)
      '10011110111101011110'

    """
    base, table = _check_base(base)
    sign = "" if number >= 0 else "-"
    number = abs(number)
    res = table[0] if number == 0 else ""
    while number:
        number, idx = divmod(number, base)
        res = table[idx] + res
    return str(sign + res)


def str2int(src, base=_DEFAULT_BASE):
    """Return the integer decoded from a string representation using a base.

    :param base: The base.
    :type base: Either an integer or a string with a custom table.

    Examples::

      >>> str2int('ffff', 16)
      65535

      >>> str2int('1c', 16) == int('1c', 16)
      True

      >>> base = 'merchise'
      >>> number = 65110208921
      >>> str2int(int2str(number, base), base) == number
      False

      >>> base = 32
      >>> str2int(int2str(number, base), base) == number
      True

    """
    base, table = _check_base(base)
    if src.startswith("-"):
        sign = -1
        i = 1
    else:
        sign = 1
        i = 0
    res = 0
    while i < len(src):
        res *= base
        res += table.index(src[i])
        i += 1
    return sign * res


class BaseConvertor:
    """Base class that implements conversion algorithms based on a simple
    lookup table and a bit mask.

    Derived classes *must* provide a `table` attribute with the table of
    digits to use.

    """

    @classmethod
    def inttobase(cls, num):
        """Converts an integer to a base representation using the class' table.

        """
        return int2str(num, base=cls.table)

    @classmethod
    def basetoint(cls, istr):
        """Converts a base representation to a integer using the class' table.

        """
        table = cls.table
        if cls.case_insensitive:
            table = table.lower()
        return str2int(istr, base=table)


class B32(BaseConvertor):
    """Handles base-32 conversions.

    In base 32, each 5-bits chunks are represented by a single "digit". Digits
    comprises all symbols in 0..9 and a..v.

        >>> B32.inttobase(32) == '10'
        True

        >>> B32.basetoint('10')
        32

    """

    table = "0123456789abcdefghijklmnopqrstuv"
    case_insensitive = True


class B64(BaseConvertor):
    """Handles [a kind of] base 64 conversions.

    This **is not standard base64**, but a reference-friendly base 64 to help
    the use case of generating a short reference.

    In base 64, each 6-bits chunks are represented by a single "digit".
    Digits comprises all symbols in 0..9, a..z, A..Z and the three symbols:
    `()[`.

        >>> B64.inttobase(64) == '10'
        True

        >>> B64.basetoint('10')
        64

   .. warning::

      In this base, letters **are** case sensitive::

          >>> B64.basetoint('a')
          10

          >>> B64.basetoint('A')
          36

    """

    table = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXZ()["
    case_insensitive = False


class B64symbolic(B64):
    """Same as B64 but uses no capital letters and lots of symbols."""

    table = "0123456789abcdefghijklmnopqrstuvwxyz" ":;><,._=!@#$^*/?\\{}%`|\"()[~'"
    case_insensitive = True



import numbers
from .meta import Quantity, Dimension

class Length:
    m: Quantity
    metre: Quantity
    km: Quantity
    kilometer: Quantity
    mm: Quantity
    millimeter: Quantity
    nm: Quantity
    nanometer: Quantity

L = Length

class Time:
    s: Quantity
    second: Quantity
    ms: Quantity
    millisecond: Quantity
    ns: Quantity
    nanosecond: Quantity
    minute: Quantity
    hour: Quantity

T = Time

class Mass:
    kg: Quantity
    kilogram: Quantity
    gram: Quantity

M = Mass

class ElectricCurrent:
    A: Quantity
    ampere: Quantity
    milliampere: Quantity

I = ElectricCurrent

class Temperature:
    K: Quantity
    kelvin: Quantity
    @classmethod
    def from_celcius(cls, val: numbers.Real) -> Quantity: ...
    @classmethod
    def from_fahrenheit(cls, val: numbers.Real) -> Quantity: ...

O = Temperature

class Substance:
    mol: Quantity
    mole: Quantity

N = Substance

class Luminosity:
    candela: Quantity

J = Luminosity

Area: Dimension
Volumen: Dimension

Frequency: Dimension
Force: Dimension

Presure: Dimension

Velocity: Dimension
Acceleration: Dimension



from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    ContextManager,
    TypeVar,
)
from typing_extensions import Protocol

from xotl.tools.symbols import Unset

# def adapt_exception(value: Optional[Union[Tuple[Type[KeyError], str], str, int]], **kwargs) -> Optional[KeyError]: ...
Getter = Callable[[Any], Callable[[str, Any], Any]]
Setter = Callable[[Any], Callable[[str, Any], None]]

def dict_merge(*dicts: Dict, **others) -> Dict: ...
def extract_attrs(obj: Any, *names: str, default: Any = ...) -> Tuple[Any, ...]: ...
def fulldir(obj: Dict[Any, Any]) -> Set[str]: ...
def get_first_of(
    source: Any, *keys: str, default: Any = ..., pred: Callable[[Any], bool] = ...
) -> Optional[Any]: ...
def get_traverser(
    *paths: str, default: Any = ..., sep: str = ..., getter: Getter = ...
) -> Callable: ...
def traverse(
    obj: Any,
    path: str,
    default: Any = Unset,
    sep: str = ".",
    getter: Optional[Getter] = None,
) -> Any: ...
def import_object(
    name: Union[object, str],
    package: str = ...,
    sep: str = ...,
    default: Any = ...,
    **kwargs
) -> Any: ...
def iterate_over(source: Any, *keys) -> Iterator[Any]: ...
def pop_first_of(source: Any, *keys: str, default: Any = ...) -> Any: ...
def save_attributes(
    obj: Any, *attrs: str, getter: Getter = ..., setter: Setter = ...
) -> ContextManager: ...
def smart_copy(*args, defaults=...) -> Dict[str, Any]: ...
def smart_getter(obj: Any, strict: bool = ...) -> Callable[[str, Any], Any]: ...
def smart_getter_and_deleter(obj: Any) -> Callable[[str, Any], Any]: ...
def smart_setter(obj: Any) -> Callable[[str, Any], None]: ...
def temp_attributes(
    obj: Any, attrs: Dict[Any, Any], getter: Getter = ..., setter: Setter = ...
) -> ContextManager: ...
def setdefaultattr(obj: Any, name: str, value: Any) -> Any: ...

class lazy:
    def __init__(self, value: Any, *args, **kawrgs) -> None: ...
    def __call__(self) -> Any: ...

class classproperty(property): ...
class staticproperty(property): ...
class memoized_property(property): ...

def copy_class(
    cls: Type,
    meta: Type = None,
    ignores: Sequence[str] = None,
    new_attrs: Mapping[str, Any] = None,
    new_name: str = None,
) -> Type: ...

B_co = TypeVar("B_co", covariant=True)

class _FinalSubclassEnum(Protocol[B_co]):
    def __getattr__(self, clsname: str) -> Type[B_co]: ...
    __members__: Mapping[str, Type[B_co]]

FinalSubclassEnumeration: _FinalSubclassEnum

class DelegatedAttribute:
    def __init__(
        self, target_name: str, delegated_attr: str, default: Any = Unset
    ) -> None: ...
    def __get__(self, instance, owner) -> Any: ...

def delegator(
    attribute: str, attrs_map: Mapping[str, str], metaclass: type = type
) -> type: ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extensions to standard library `mimetype`:mod:.

This module reexport all functions the *current* version of Python.

.. versionadded:: 1.8.4

"""


import mimetypes as _stdlib
from mimetypes import *  # noqa

from xotl.tools.symbols import Unset


def guess_type(url, strict=True, default=(None, None)):
    """Guess the type of a file based on its filename or URL, given by url.

    This is the same as `mimetypes.guess_type`:func: with the addition of the
    `default` keyword

    """
    type, encoding = _stdlib.guess_type(url, strict=strict)
    if default is not Unset and type is None:
        type, encoding = default
    return type, encoding



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest


class TestCLisp(unittest.TestCase):
    def test_basic_coercers(self):
        from xotl.tools.values import (
            identity_coerce,
            void_coerce,
            coercer,
            vouch,
            t,
            int_coerce,
            float_coerce,
            create_int_range_coerce,
            istype,
            typecast,
            iterable,
            mapping,
            create_unique_member_coerce,
            nil,
        )

        d = {"1": 2, 3.0: "4", 5.0 + 0j: 7.3 + 0j, 1: "2"}
        s = {1, "2", 3.0, "1"}
        l = [1, "2", 3.0, "1", "x10"]
        number_types = (int, float, complex)
        mc = mapping(int_coerce, float_coerce)
        uint_coerce = create_unique_member_coerce(int_coerce, d)
        mcu = mapping(uint_coerce, float_coerce)
        ic = iterable(int_coerce)
        age_coerce = create_int_range_coerce(0, 100)
        text_coerce = coercer(str)
        isnumber = istype(number_types)
        numbercast = typecast(number_types)
        # TODO: don't use isinstance
        self.assertEqual(
            all(
                isinstance(c, coercer)
                for c in (
                    mc,
                    mcu,
                    uint_coerce,
                    ic,
                    age_coerce,
                    text_coerce,
                    identity_coerce,
                    void_coerce,
                    int_coerce,
                    float_coerce,
                )
            ),
            True,
        )
        self.assertEqual(mc(dict(d)), {1: 2.0, 3: 4.0, 5: 7.3})
        self.assertIs(mcu(d), nil)
        self.assertEqual(mcu.scope, ({"1": 2}, uint_coerce))
        self.assertEqual(ic(s), {1, 2, 3})
        self.assertIs(ic(l), nil)
        self.assertIs(ic.scope, l[-1])
        self.assertEqual(l, [1, 2, 3, 1, "x10"])
        self.assertIs(age_coerce(80), 80)
        self.assertFalse(t(age_coerce(120)))
        self.assertIs(vouch(age_coerce, 80), 80)
        with self.assertRaises(TypeError):
            vouch(age_coerce, 120)
        self.assertIs(isnumber(5), 5)
        self.assertIs(isnumber(5.1), 5.1)
        with self.assertRaises(TypeError):
            vouch(isnumber, "5.1")
        self.assertIs(numbercast(5), 5)
        self.assertIs(numbercast(5.1), 5.1)
        self.assertEqual(numbercast("5.1"), 5.1)
        self.assertIs(numbercast.scope, float)

    def test_compound_coercers(self):
        from xotl.tools.values import (
            coercer,
            compose,
            some,
            combo,
            iterable,
            typecast,
            int_coerce,
            float_coerce,
            nil,
        )

        isstr = coercer(str)
        strcast = typecast(str)
        toint = compose(isstr, int_coerce)
        isint = some(isstr, int_coerce)
        hyphenjoin = coercer(lambda arg: "-".join(arg))
        intjoin = compose(iterable(strcast), hyphenjoin)
        cb = combo(strcast, int_coerce, float_coerce)
        self.assertEqual(toint("10"), 10)
        self.assertIs(toint(10), nil)
        self.assertEqual(toint.scope, (10, isstr))
        self.assertEqual(isint("10"), "10")
        self.assertEqual(isint.scope, isstr)
        self.assertEqual(isint(10), 10)
        self.assertEqual(isint.scope, int_coerce)
        self.assertEqual(intjoin(2 * i + 1 for i in range(5)), "1-3-5-7-9")
        self.assertEqual(cb([1, "2.0", 3, 4]), ["1", 2, 3.0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest
from xotl.tools.decorator import assignment_operator
from xotl.tools.decorator.meta import decorator


class TestAssignable(unittest.TestCase):
    def test_inline_expression(self):
        @assignment_operator()
        def test(name, *args):
            return name * (len(args) + 1)

        self.assertEqual("aaa", test("a", 1, 2))

    def test_assignment(self):
        @assignment_operator()
        def test(name, *args):
            return name * (len(args) + 1)

        b = test(1, 2, 4)
        self.assertEqual("bbbb", b)

    def test_regression_inline(self):
        @assignment_operator(maybe_inline=True)
        def test(name, *args):
            if name:
                return name * (len(args) + 1)
            else:
                return None

        self.assertIs(None, test("a", 1, 2))

    def test_regression_on_block(self):
        @assignment_operator(maybe_inline=True)
        def union(name, *args):
            return (name,) + args

        for which in (union(1, 2),):
            self.assertEqual((None, 1, 2), which)

    def test_argsless_decorator(self):
        @decorator
        def log(func, fmt="Calling function %s"):
            def inner(*args, **kwargs):
                print(fmt % func.__name__)
                return func(*args, **kwargs)

            return inner

        @log
        def hello(msg="Hi"):
            print(msg)

        @log()
        def hi(msg="Hello"):
            print(msg)

        hi()
        hello()
        pass

    def test_returning_argless(self):
        @decorator
        def plus2(func, value=1):
            def inner(*args):
                return func(*args) + value

            return inner

        @plus2
        def ident2(val):
            return val

        @plus2()
        def ident3(val):
            return val

        self.assertEqual(ident2(10), 11)
        self.assertEqual(ident3(10), 11)


class RegressionTests(unittest.TestCase):
    def test_with_kwargs(self):
        """When passing a function as first positional argument, kwargs should
        be tested empty.

        """
        from xotl.tools.future.functools import partial

        @decorator
        def ditmoi(target, *args, **kwargs):
            return partial(target, *args, **kwargs)

        def badguy(n):
            return n

        @ditmoi(badguy, b=1)
        def foobar(n, *args, **kw):
            return n

        self.assertEqual(badguy, foobar(1))


class Memoizations(unittest.TestCase):
    def test_memoized_property(self):
        from xotl.tools.future.inspect import getattr_static
        from xotl.tools.objects import memoized_property

        class Foobar:
            @memoized_property
            def prop(self):
                return self

            with self.assertRaises(AttributeError):

                @prop.setter
                def prop(self, value):
                    pass

            with self.assertRaises(AttributeError):

                @prop.deleter
                def prop(self, value):
                    pass

        foo = Foobar()
        self.assertNotEquals(getattr_static(foo, "prop"), foo)
        self.assertIs(foo.prop, foo)
        self.assertIs(getattr_static(foo, "prop"), foo)
        # After the first invocation, the static attr is the result.
        Foobar.prop.reset(foo)
        self.assertNotEquals(getattr_static(foo, "prop"), foo)


class ConstantBags(unittest.TestCase):
    def test_constant_bags_decorator(self):
        from xotl.tools.decorator import constant_bagger as typify

        def func(**kwds):
            return kwds

        bag = func(ONE=1, TWO=2)

        @typify(ONE=1, TWO=2)
        def BAG(**kwds):
            return kwds

        self.assertIs(type(BAG), type)
        self.assertIn("ONE", bag)
        self.assertEqual(bag["ONE"], BAG.ONE)
        self.assertEqual(BAG.TWO, 2 * BAG.ONE)
        with self.assertRaises(AttributeError):
            self.assertEqual(bag.TWO, 2 * bag.ONE)
        with self.assertRaises(TypeError):
            self.assertEqual(BAG["TWO"], 2 * BAG["ONE"])
        with self.assertRaises(AttributeError):
            self.assertEqual(BAG.THREE, 3)
        self.assertIs(BAG(THREE=3), BAG)
        self.assertEqual(BAG.THREE, 3)


if __name__ == "__main__":
    unittest.main(verbosity=2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""General security tools.

Adds the ability to generate new passwords using a source pass-phrase and a
secury strong level.

"""


__all__ = (
    "PASS_PHRASE_LEVEL_BASIC",
    "PASS_PHRASE_LEVEL_MAPPED",
    "PASS_PHRASE_LEVEL_MAPPED_MIXED",
    "PASS_PHRASE_LEVEL_MAPPED_DATED",
    "PASS_PHRASE_LEVEL_STRICT",
    "generate_password",
)


#: The most basic level (less ) for the password generation.
PASS_PHRASE_LEVEL_BASIC = 0

#: A level for simply mapping of several chars.
PASS_PHRASE_LEVEL_MAPPED = 1

#: Another "stronger" mapping level.
PASS_PHRASE_LEVEL_MAPPED_MIXED = 2

#: Appends the year after mapping.
PASS_PHRASE_LEVEL_MAPPED_DATED = 3

#: Totally scramble the result, making very hard to predict the result.
PASS_PHRASE_LEVEL_STRICT = 4

#: The default level for `generate_password`:func:
DEFAULT_PASS_PHRASE_LEVEL = PASS_PHRASE_LEVEL_MAPPED_DATED


#: A mapping from names to standards levels.  You may use these strings as
#: arguments for `level` in `generate_password`:func:.
PASS_LEVEL_NAME_MAPPING = {
    "basic": PASS_PHRASE_LEVEL_BASIC,
    "mapped": PASS_PHRASE_LEVEL_MAPPED,
    "mixed": PASS_PHRASE_LEVEL_MAPPED_MIXED,
    "dated": PASS_PHRASE_LEVEL_MAPPED_DATED,
    "random": PASS_PHRASE_LEVEL_STRICT,
}


BASIC_PASSWORD_SIZE = 4  # bytes

#: An upper limit for generated password length.
MAX_PASSWORD_SIZE = 512


SAMPLE = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"


def _normalize_level(level):
    """Normalize the `level` argument.

    If passed a string, it must be a key in `PASS_LEVEL_NAME_MAPPING`:obj:.
    Otherwise it must be a valid level number.

    """
    if isinstance(level, str):
        return PASS_LEVEL_NAME_MAPPING[level]
    else:
        return level


def generate_password(pass_phrase, level=DEFAULT_PASS_PHRASE_LEVEL):
    """Generate a password from a source `pass-phrase` and a security `level`.

    :param pass_phrase: String pass-phrase to be used as base of password
                        generation process.

    :param level: Numerical security level (the bigger the more secure, but
                  don't exaggerate!).

    When `pass_phrase` is a valid string, `level` means a generation method.
    Each level implies all other with an inferior numerical value.

    There are several definitions with numerical values for `level` (0-4):

    `PASS_PHRASE_LEVEL_BASIC`:data:

        Generate the same pass-phrase, just removing invalid characters and
        converting the result to lower-case.

    `PASS_PHRASE_LEVEL_MAPPED`:data:

        Replace some characters with new values: ``'e'->'3'``, ``'i'->'1'``,
        ``'o'->'0'``, ``'s'->'5'``.

    `PASS_PHRASE_LEVEL_MAPPED_MIXED`:data:

        Consonants characters before 'M' (included) are converted to
        upper-case, all other are kept lower-case.

    `PASS_PHRASE_LEVEL_MAPPED_DATED`:data:

        Adds a suffix with the year of current date ("<YYYY>").

    `PASS_PHRASE_LEVEL_STRICT`:data:

        Randomly scramble previous result until unbreakable strong password is
        obtained.

    If `pass_phrase` is ``None`` or an empty string, generate a "secure salt"
    (a password not based in a source pass-phrase).  A "secure salt" is
    generated by scrambling the concatenation of a random phrases from the
    alphanumeric vocabulary.

    Returned password size is ``4*level`` except when a `pass-phrase` is given
    for `level` <= 4 where depend on the count of valid characters of
    `pass-phrase` argument, although minimum required is warranted.  When
    `pass-phrase` is ``None`` for `level` zero or negative, size ``4`` is
    assumed.  First four levels are considered weak.

    Maximum size is defined in the :data:`MAX_PASSWORD_SIZE` constant.

    Default level is :data:`PASS_PHRASE_LEVEL_MAPPED_DATED` when using a
    pass-phrase.

    """
    from random import sample, randint
    from xotl.tools.string import slugify

    level = _normalize_level(level)
    size = MAX_PASSWORD_SIZE + 1  # means, return all calculated
    required = min(max(level, 1) * BASIC_PASSWORD_SIZE, MAX_PASSWORD_SIZE)
    if pass_phrase:
        # PASS_PHRASE_LEVEL_BASIC
        res = slugify(pass_phrase, "", invalid_chars="_")
        if level >= PASS_PHRASE_LEVEL_MAPPED:
            for (old, new) in ("e3", "i1", "o0", "s5"):
                res = res.replace(old, new)
        if level >= PASS_PHRASE_LEVEL_MAPPED_MIXED:
            for new in "BCDFGHJKLM":
                old = new.lower()
                res = res.replace(old, new)
        if level >= PASS_PHRASE_LEVEL_MAPPED_DATED:
            from datetime import datetime

            today = datetime.today()
            res += today.strftime("%Y")
        if level >= PASS_PHRASE_LEVEL_STRICT:
            size = required
    else:
        size = required
        count = randint(BASIC_PASSWORD_SIZE, 2 * BASIC_PASSWORD_SIZE)
        res = "".join(sample(SAMPLE, count))
    if size <= MAX_PASSWORD_SIZE:
        if len(res) < size:
            needed = (size - len(res)) // BASIC_PASSWORD_SIZE + 1
            res += generate_password(None, needed)
        res = "".join(sample(res, size))
    return res[:size]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extensions to Python's `threading` module.

You may use it as drop-in replacement of ``threading``. Although we don't
document all items here.  Refer to `threading`:mod: documentation.

"""

from threading import *  # noqa
import threading as _stdlib  # noqa

from threading import Event, Thread, RLock, Timer


def async_call(func, args=None, kwargs=None, callback=None, onerror=None):
    """Executes a function asynchronously.

    The function receives the given positional and keyword arguments

    If `callback` is provided, it is called with a single positional argument:
    the result of calling `func(*args, **kwargs)`.

    If the called function ends with an exception and `onerror` is provided, it
    is called with the exception object.

    :returns: An event object that gets signalled when the function ends its
              execution whether normally or with an error.

    :rtype: `Event`:class:

    """
    event = Event()
    event.clear()
    if not args:
        args = ()
    if not kwargs:
        kwargs = {}

    def async_():
        try:
            result = func(*args, **kwargs)
            if callback:
                callback(result)
        except Exception as error:
            if onerror:
                onerror(error)
        finally:
            event.set()

    thread = Thread(target=async_)
    thread.setDaemon(True)  # XXX: Why?
    thread.start()
    return event


class _SyncronizedCaller:
    """Protected to be used in `sync_call`:func:"""

    def __init__(self, pooling=0.005):
        self.lock = RLock()
        self._not_bailed = True
        self.pooling = pooling

    def __call__(self, funcs, callback, timeout=None):
        def _syncronized_callback(result):
            with self.lock:
                if self._not_bailed:
                    callback(result)

        events, threads = [], []
        for which in funcs:
            event, thread = async_call(which, callback=_syncronized_callback)
            events.append(event)
            threads.append(thread)
        if timeout:

            def set_all_events():
                with self.lock:
                    self._not_bailed = False
                for e in events:
                    e.set()

            timer = Timer(timeout, set_all_events)
            timer.start()
        while events:
            terminated = []
            for event in events:
                flag = event.wait(self.pooling)
                if flag:
                    terminated.append(event)
            for e in terminated:
                events.remove(e)
        if timeout:
            timer.cancel()


def sync_call(funcs, callback, timeout=None):
    """Calls several functions, each one in it's own thread.

    Waits for all to end.

    Each time a function ends the `callback` is called (wrapped in a lock to
    avoid race conditions) with the result of the as a single positional
    argument.

    If `timeout` is not None it sould be a float number indicading the seconds
    to wait before aborting. Functions that terminated before the timeout will
    have called `callback`, but those that are still working will be ignored.

    .. todo:: Abort the execution of a thread.

    :param funcs: A sequences of callables that receive no arguments.

    """
    sync_caller = _SyncronizedCaller()
    sync_caller(funcs, callback, timeout)


from threading import __all__  # noqa

__all__ = list(__all__) + ["async_call", "sync_call"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""File system utilities.

This module contains file-system utilities that could have side-effects. For
path-handling functions that have no side-effects look at
`xotl.tools.fs.path`:mod:.

"""


import sys
import os
from re import compile as _rcompile
from xotl.tools.fs.path import normalize_path


re_magic = _rcompile("[*?[]")
has_magic = lambda s: re_magic.search(s) is not None


def _get_regex(pattern=None, regex_pattern=None, shell_pattern=None):
    from functools import reduce
    import fnmatch
    from xotl.tools.params import check_count

    arg_count = reduce(
        lambda count, p: count + (1 if p is not None else 0),
        (pattern, regex_pattern, shell_pattern),
        0,
    )
    check_count(arg_count, 0, 1, caller="_get_regex")  # XXX: WTF?!
    if arg_count == 1:
        if pattern is not None:
            if pattern.startswith("(?") or pattern.startswith("^(?"):
                regex_pattern = pattern
            else:
                shell_pattern = pattern
        return _rcompile(regex_pattern or fnmatch.translate(shell_pattern))
    elif arg_count == 0:
        return None


def iter_files(
    top=".",
    pattern=None,
    regex_pattern=None,
    shell_pattern=None,
    followlinks=False,
    maxdepth=None,
):
    """Iterate filenames recursively.

    :param top: The top directory for recurse into.
    :param pattern: A pattern of the files you want to get from the iterator.
                    It should be a string. If it starts with "(?" it will be
                    regarded as a regular expression, otherwise a shell
                    pattern.

    :param regex_pattern: An *alternative* to `pattern`. This will always be
                          regarded as a regular expression.

    :param shell_pattern: An *alternative* to `pattern`. This should be a
                          shell pattern.

    :param followlinks: The same meaning that in `os.walk`.

    :param maxdepth: Only files above this level will be yielded. If None, no
                     limit is placed.

    .. warning:: It's an error to pass more than pattern argument.

    .. versionchanged:: 1.2.1 Added parameters `followlinks` and `maxdepth`.

    """
    regex = _get_regex(pattern, regex_pattern, shell_pattern)
    depth = 0
    for dirpath, _dirs, filenames in os.walk(
        normalize_path(top), topdown=True, followlinks=followlinks
    ):
        for filename in filenames:
            path = os.path.join(dirpath, filename)
            if (regex is None) or regex.search(path):
                yield path
        if maxdepth is not None:
            depth += 1
            if depth >= maxdepth:
                _dirs[:] = []


# ------------------------------ iter_dict_files ------------------------------
_REGEX_PYTHON_PACKAGE = _rcompile(
    r"^(?P<dir>.+(?=/)/)?"
    r"(?P<packagename>[^/_-]+?)"
    r"([-_][Vv]?(?P<version>\d+([.-_]\w+)*))?"
    r"(?P<ext>[.](tar[.](gz|bz2)|zip|egg|tgz))$"
)

_REGEX_DEFAULT_ALLFILES = _rcompile(
    r"^(?P<dir>.+(?=/)/)?" r"(?P<filename>[^/]+?)" r"([.](?P<ext>[^.]+))?$"
)


def iter_dict_files(top=".", regex=None, wrong=None, followlinks=False):
    """
    Iterate filenames recursively.

    :param top: The top directory for recurse into.

    :param regex: Regular expression with group definitions to match.

    :param wrong: A key to store full name of not matching files.

    :param followlinks: The same meaning that in `os.walk`.

    .. versionadded:: 1.2.0

    .. versionchanged:: 1.2.1 Added parameter `followlinks`.

    """
    if regex:
        if isinstance(regex, str):
            regex = _rcompile(regex)
    else:
        regex = _REGEX_DEFAULT_ALLFILES
    for dirpath, _dirs, filenames in os.walk(
        normalize_path(top), followlinks=followlinks
    ):
        for filename in filenames:
            path = os.path.join(dirpath, filename)
            match = regex.match(path)
            if match:
                yield match.groupdict()
            elif wrong is not None:
                yield {wrong: path}


def iter_dirs(top=".", pattern=None, regex_pattern=None, shell_pattern=None):
    """
    Iterate directories recursively.

    The params have analagous meaning that in `iter_files`:func: and the same
    restrictions.

    """
    regex = _get_regex(pattern, regex_pattern, shell_pattern)
    for path, _dirs, _files in os.walk(normalize_path(top)):
        if (regex is None) or regex.search(path):
            yield path


def rmdirs(
    top=".",
    pattern=None,
    regex_pattern=None,
    shell_pattern=None,
    exclude=None,
    confirm=None,
):
    """Removes all empty dirs at `top`.

    :param top: The top directory to recurse into.

    :param pattern: A pattern of the dirs you want to remove.
                    It should be a string. If it starts with "(?" it will be
                    regarded as a regular expression, otherwise a shell
                    pattern.

    :param exclude: A pattern of the dirs you DON'T want to remove.  It should
                    be a string. If it starts with "(?" it will be regarded as
                    a regular expression, otherwise a shell pattern. This is a
                    simple commodity to have you not to negate complex
                    patterns.

    :param regex_pattern: An *alternative* to `pattern`. This will always be
                          regarded as a regular expression.

    :param shell_pattern: An *alternative* to `pattern`. This should be a
                          shell pattern.

    :param confirm: A callable that accepts a single argument, which is
                    the path of the directory to be deleted. `confirm`
                    should return True to allow the directory to be
                    deleted. If `confirm` is None, then all matched dirs
                    are deleted.

    .. note:: In order to avoid common mistakes we won't attempt to
              remove mount points.

    .. versionadded:: 1.1.3

    """
    regex = _get_regex(pattern, regex_pattern, shell_pattern)
    exclude = _get_regex(exclude)
    if confirm is None:
        confirm = lambda _: True
    for path, _dirs, _files in os.walk(normalize_path(top)):
        # XXX: Make clearest next condition
        if (
            (regex is None or regex.search(path))
            and (exclude is None or not exclude.search(path))
            and not _dirs
            and not _files
            and confirm(path)
            and not os.path.ismount(path)
        ):
            os.rmdir(path)


def regex_rename(top, pattern, repl, maxdepth=None):
    """Rename files recursively using regular expressions substitution.

    :param top: The top directory to start walking.

    :param pattern: A regular expression pattern.  Files whose fullname
           (including the path) match the expression will be renamed.

    :param repl: String to use as replacement. You may use backreferences as
           documented in python's ``re.sub`` function.

    :param maxdepth: Only walk files up to this level.  If None, walk all
           files.

       .. versionadded:: 1.2.1

    """
    from re import subn as _re_subn

    if isinstance(pattern, str):
        pattern = _rcompile(pattern)
    depth = 0
    for path, _dirs, files in os.walk(top):
        for item in files:
            new_file, count = _re_subn(pattern, repl, item)
            if count > 0:
                old = os.path.join(path, item)
                new = os.path.join(path, new_file)
                os.rename(old, new)
        if maxdepth is not None:
            depth += 1
            if depth >= maxdepth:
                _dirs[:] = []


filter_not_hidden = lambda path, _st: (path[0] != ".") and ("/." not in path)
filter_false = lambda path, stat_info: False


def get_regex_filter(regex):
    """Return a filter for "walk" based on a regular expression."""
    if isinstance(regex, str):
        regex = _rcompile(regex)

    def _filter(path, stat_info):
        return regex.match(os.path.basename(path)) is not None

    return _filter


def get_wildcard_filter(pattern):
    """Return a filter for "walk" based on a wildcard pattern a la fnmatch."""
    regex = _get_regex(pattern)

    def _filter(path, stat_info):
        return regex.match(os.path.basename(path)) is not None

    return _filter


def get_mime_filter(mime_start):
    import mimetypes

    def _filter(path, stat_info):
        t = mimetypes.guess_type(path)[0]
        return t.startswith(mime_start) if t else False

    return _filter


def nice_size(size):
    """Formats `size` to a nice human-friendly format by appending one of `Kilo`,
    `Mega`, `Giga`, `Tera`, `Peta`, or `Eta` suffix.

    """
    tails = " KMGTPE"
    order, highest = 0, len(tails) - 1
    while (size >= 1024) and (order < highest):
        size /= 1024
        order += 1
    res = ("%.2f" % size).rstrip("0").rstrip(".")
    return "%s%s" % (res, tails[order])


def stat(path):
    """
    Return file or file system status.

    This is the same as the function ``os.stat`` but raises no error.
    """
    try:
        return os.stat(path)
    except os.error:
        return None


def lstat(path):
    """Same as `os.lstat`, but raises no error."""
    try:
        return os.lstat(path)
    except os.error:
        return None


def set_stat(fname, stat_info):
    os.chmod(fname, stat_info.st_mode)
    os.chown(fname, stat_info.st_uid, stat_info.st_gid)
    os.utime(fname, (stat_info.st_atime, stat_info.st_mtime))


def read_file(path):
    """Read a full file content and return an string."""
    try:
        with open(path, "r") as f:
            return f.read()
    except OSError:
        return ""


def listdir(path):
    """Same as ``os.listdir`` but normalizes `path` and raises no error."""
    try:
        return os.listdir(normalize_path(path))
    except os.error:
        return []


def _list_magic(dirname, pattern):
    re = _get_regex(pattern)
    for name in listdir(dirname or os.curdir):
        if re.match(name):
            full = os.path.join(dirname, name)
            yield full, lstat(full)


def _list_one(fname):
    st = lstat(fname)
    if st:
        yield fname, st


def _list(pattern):
    from stat import S_ISDIR as _ISDIR

    if has_magic(pattern):
        head, tail = os.path.split(pattern)
        for dirname, st in _list(head):
            if _ISDIR(st.st_mode):
                if has_magic(tail):
                    items = _list_magic(dirname, tail)
                elif tail:
                    items = _list_one(os.path.join(dirname, tail))
                else:
                    items = ((dirname, st),)
            for item in items:
                yield item
    elif pattern:
        for item in _list_one(pattern):
            yield item
    else:
        yield ("", lstat(os.curdir))


def imap(func, pattern):
    r"""Yields `func(file_0, stat_0)`, `func(file_1, stat_1)`, ... for each dir
    path. The `pattern` may contain:

    - Simple shell-style wild-cards à la `fnmatch`.

    - Regex if pattern starts with '(?'. Expressions must be valid, as
      in "(?:[^.].*)$" or "(?i).*\.jpe?g$". Remember to add the end mark '$'
      if needed.

    """
    for item, st in _list(pattern):
        res = func(item, st)
        if res is not None:
            yield res


def walk_up(start, sentinel):
    """Given a `start` directory walk-up the file system tree until either the
    FS root is reached or the `sentinel` is found.

    The `sentinel` must be a string containing the file name to be found.

    .. warning:: If `sentinel` is an absolute path that exists this will return
       `start`, no matter what `start` is (in windows this could be even
       different drives).

    If `start` path exists but is not a directory an OSError is raised.

    """
    from os.path import abspath, exists, isdir, join, dirname

    current = abspath(start)
    if not exists(current) or not isdir(current):
        raise OSError('Invalid directory "%s"' % current)
    previouspath = None
    found = False
    while not found and current is not previouspath:
        clue = join(current, sentinel)
        if exists(clue):
            found = True
        else:
            previouspath = current
            current = dirname(current)
    return current if found else None


from os import makedirs


def ensure_filename(filename, yields=False):
    """Ensures the existence of a file with a given filename.

    If the filename is taken and is not pointing to a file (or a link to a
    file) an OSError is raised.  If `exist_ok` is False the filename must not
    be taken; an OSError is raised otherwise.

    The function creates all directories if needed. See `makedirs`:func: for
    restrictions.

    If `yields` is True, returns the file object.  This way you may open a
    file for writing like this::

      with ensure_filename('/tmp/good-name-87.txt', yields=True) as fh:
          fh.write('Do it!')

    The file is open in mode 'w+b'.

    .. versionadded:: 1.6.1  Added parameter `yield`.

    """
    if not os.path.exists(filename):
        filename = normalize_path(filename)
        dirname = os.path.dirname(filename)
        makedirs(dirname, exist_ok=True)
        # TODO: Better hanlding of mode for reading/writing.
        fh = open(filename, "w+b")
        if not yields:
            fh.close()
        else:
            return fh
    else:
        if not os.path.isfile(filename):
            raise OSError("Expected a file but another thing is found '%s'" % filename)


def concatfiles(*files):
    """Concat several files to a single one.

    Each positional argument must be either:

    - a file-like object (ready to be passed to `shutil.copyfileobj`:func:)

    - a string, the file path.

    The last positional argument is the target.  If it's file-like object it
    must be open for writing, and the caller is the responsible for closing
    it.

    Alternatively if there are only two positional arguments and the first is
    a collection, the sources will be the members of the first argument.

    """
    import shutil
    from xotl.tools.values.simple import force_iterable_coerce
    from xotl.tools.params import check_count

    check_count(files, 2, caller="concatfiles")
    if len(files) == 2:
        files, target = force_iterable_coerce(files[0]), files[1]
    else:
        files, target = files[:-1], files[-1]
    if isinstance(target, str):
        target, opened = open(target, "wb"), True
    else:
        opened = False
    try:
        for f in files:
            if isinstance(f, str):
                fh = open(f, "rb")
                closefh = True
            else:
                fh = f
                closefh = False
            try:
                shutil.copyfileobj(fh, target)
            finally:
                if closefh:
                    fh.close()
    finally:
        if opened:
            target.close()


del sys



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""CSV parsing and writing extensions.

This module is an extension of `csv`:mod: Python standard module, it provides
classes and tools that assist in the reading and writing of `Comma Separated
Value (CSV)`:abbr: files, and implements the interface described by PEP
`305`:pep:.

"""


from csv import *  # noqa
from csv import unix_dialect
import csv as _stdlib  # noqa


#: Define 'unix dialect' as our base default for inheritance.
DefaultDialect = unix_dialect

reader = _stdlib.reader


def parse(data, *dialect, **options):
    r"""Parse `data` into a sheet.

    This function has the exact parameters protocol as `~csv.reader`:func:\ ::

      parse(data [, dialect='excel'] [, optional keyword options])

    :param data: Can be any object that returns a line of input for each
           iteration, such as a file object or a list.

    :param dialect: An optional parameter can be given which is used to define
           a set of parameters specific to a particular CSV dialect.  It may
           be an instance of a subclass of the `~csv.Dialect`:class: class or
           one of the strings returned by the `~csv.list_dialects`:func:
           function.

    The other optional keyword arguments can be given to override
    individual formatting parameters in the current `dialect`.

    When reading a value, `csv`:mod: for Python version 2 doesn't accept
    unicode text, so given data lines are forced to be `str`:class: values
    before processed by `~csv.reader`:func:.  Each cell is converted to
    unicode text after read.

    :returns: The parsed matrix.

    A short usage example::

      >>> from xotl.tools.future import csv
      >>> with open('test.csv', newline='') as data:
      ...     matrix = csv.parse(data)
      ...     for row in matrix:
      ...         print(', '.join(row))
      Last name, First Name
      van Rossum, Guido
      Stallman, Richard

    """
    string_force = text_force = str  # They were different in Python 2.
    rows = reader((string_force(line) for line in data), *dialect, **options)
    return [[text_force(cell) for cell in row] for row in rows]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Facilities to work with `concrete numbers`_.

The name `dim`:mod: is a short of dimension.  We borrow it from the topic
"dimensional analysis", even though the scope of this module is less
ambitious.

This module is divided in two major parts:

- `xotl.tools.dim.meta`:mod: which allows to define almost any kind of
  quantity decorated with a unit.

- Other modules ``xotl.tools.dim.*`` which contains applications of the
  definitions in `~xotl.tools.dim.meta`:mod:.  In particular,
  `xotl.tools.dim.base`:mod: contains the `base quantities`_ for the
  `International System of Quantities`_.


.. _concrete numbers: https://en.wikipedia.org/wiki/Concrete_number

.. _base quantities: https://en.wikipedia.org/wiki/Base_quantity

.. _International System of Quantities: \
   https://en.wikipedia.org/wiki/International_System_of_Quantities

"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
'''Transition to a new namespace
-----------------------------

Since version 2.1, we're transitioning to another name: ``xotl.tools``.  This
is to align ``xoutil`` as a fundamental part of our family of projects under
the ``xotl`` namespace.  *Xotl* is a Nahuatl word which may stand for
'foundation'.  ``xoutil`` is part of the foundation of many of our projects.

Backwards compatible imports
----------------------------

Since 2.1, every module importable from ``xoutil`` is actually under the
namespace ``xotl.tools``; so importing, for instance, from
``xoutil.future.datetime`` should be updated to
`xotl.tools.future.datetime`:mod:.

Importing from ``xoutil`` will still be possible in all versions before 3.0.
You won't have to change all your imports right away.


Distribution of ``xoutil``
--------------------------

Will continue to distribute both `xotl.tools <xotl-tools-dist_>`__ and `xoutil
<xoutil-dist_>`__ (with the same codebase) for the entire 2.1.x series.  From
version 2.2.0+ will distruibute only ``xotl.tools``, but keep the backwards
import up to 3.0.

.. warning:: Don't depend on both ``xoutil`` and ``xotl.tools``.  We use the
   same codebase for both distributions; which means you'll get the same code,
   but if you install different versions you may get a crippled system.

.. _xotl-tools-dist: https://pypi.org/project/xotl.tools
.. _xoutil-dist: https://pypi.org/project/xoutil

'''

import sys
import importlib
from importlib.abc import MetaPathFinder

XOUTIL_NAMESPACE = 'xoutil.'
XOTL_TOOLS_NS = 'xotl.tools.'


class Hook(MetaPathFinder):
    def find_module(self, full_name, path=None):
        name = self._from_xoutil_to_xotl(full_name)
        if name:
            return self

    def _from_xoutil_to_xotl(self, full_name):
        if full_name.startswith(XOUTIL_NAMESPACE):
            path = full_name[len(XOUTIL_NAMESPACE):]
            return XOTL_TOOLS_NS + path
        else:
            return None

    def load_module(self, full_name):
        result = sys.modules.get(full_name, None)
        if result:
            return result
        modname = self._from_xoutil_to_xotl(full_name)
        result = None
        if modname:
            result = sys.modules.get(modname, None)
        if not result and modname:
            result = importlib.import_module(modname)
        if result:
            import warnings
            warnings.warn(
                ('Importing from xoutil ({}) is deprecated; '
                 'import from xotl.tools ({})').format(full_name, modname)
            )
            sys.modules[modname] = sys.modules[full_name] = result
            return result
        else:
            raise ImportError(modname)


# I have to put this meta path before Python's standard, because otherwise
# importing like this:
#
#     from xoutil.future.datetime import TimeSpan
#
# First imports 'xoutil.future' with our Hook, but afterwards,
# 'xoutil.future.datetime' is found by _frozen_importlib_external.PathFinder
# instead of our own Hook which maintains 100% backwards compatibility of
# imports.
sys.meta_path.insert(0, Hook())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest
from xotl.tools.bound import boundary, whenall, whenany


def fibonacci(wait=None):
    import time

    a, b = 1, 1
    while True:
        if wait:
            time.sleep(wait)
        yield a
        a, b = b, a + b


class TestBoundedWithStandardPredicates(unittest.TestCase):
    def test_times(self):
        from xotl.tools.bound import times, until

        fib8 = times(8)(fibonacci)
        # Fibonacci numbers are yielded:
        # 1 1 2 3 5 8 13 21
        self.assertEqual(fib8(), 21)

        fib8 = until(times=8)(fibonacci)
        # Fibonacci numbers are yielded:
        # 1 1 2 3 5 8 13 21
        self.assertEqual(fib8(), 21)

        fib8 = times(8)(fibonacci)

        fib8gen = fib8.generate()  # exposed bounded generator
        self.assertEqual(tuple(fib8gen), (1, 1, 2, 3, 5, 8, 13, 21))

    def test_until_error(self):
        from xotl.tools.bound import until

        d = dict(a=1, b=2, c=3, d=4)

        @until(errors=(KeyError,))
        def getall(d, *keys):
            for k in keys:
                yield d[k]

        assert d["d"] == getall(d, "a", "b", "d")
        assert d["a"] == getall(d, "a", "kkk")

        @until(errors=(ValueError,))
        def getall(d, *keys):
            for k in keys:
                yield d[k]

        with self.assertRaises(KeyError):
            getall(d, "kkk")

        @until(errors=(RuntimeError,))
        def failing():
            raise RuntimeError
            yield 1

        assert failing() is None
        assert list(failing.generate()) == []

    def test_timed(self):
        from xotl.tools.bound import timed, until

        fib10ms = timed(1 / 100)(fibonacci)
        # Since the wait time below will be larger than the allowed execution
        # (10 ms) fib1ms will only be able to yield a single value (notice
        # that `timed` always allow a cycle.)
        res = fib10ms(wait=1 / 10)
        self.assertEqual(res, 1)

        fib10ms = until(maxtime=1 / 100)(fibonacci)
        # Since the wait time below will be larger than the allowed execution
        # (10 ms) fib1ms will only be able to yield a single value (notice
        # that `timed` always allow a cycle.)
        res = fib10ms(wait=1 / 10)
        self.assertEqual(res, 1)

        # If the time boundary is too low timed will allow not allow a cycle.
        fib0ms = timed(0)(fibonacci)
        res = fib0ms()
        self.assertEqual(res, None)

    def test_accumulated(self):
        from xotl.tools.bound import until
        from xotl.tools.bound import accumulated, timed, times

        # 1 + 1 + 2 + 3 + 5 + 8 + 13 + 21 + 34 + 55 + 89 + 144 = 376
        # ^   ^        ...                                  ^
        # |   |        ...                                  |
        # 1   2   3   4   5   6    7    8    9   10   11    12    13
        # |   |        ...                                  |     |
        # V   V        ...                                  V     V
        # 1 + 1 + 2 + 3 + 5 + 8 + 13 + 21 + 34 + 55 + 89 + 144 + 233 = 609
        fib500 = accumulated(500)(fibonacci)
        self.assertEqual(fib500(), 233)
        fib500 = until(accumulate=500)(fibonacci)
        self.assertEqual(fib500(), 233)

        fib500timed = whenall(accumulated(500), timed(0))(fibonacci)
        self.assertEqual(fib500timed(), 233)

        self.assertEqual(
            tuple(fib500.generate()), (1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233)
        )

        # With .generate()  you may count
        self.assertEqual(len(tuple(fib500.generate())), 13)  # the 13th

        # Since 500 is reached at the 13th fib number, looping up to the 20th
        # number must be bigger.
        fib500at20 = whenall(accumulated(500), times(20))(fibonacci)
        self.assertGreater(fib500at20(), 233)


class TestBoundaryDefinitions(unittest.TestCase):
    def test_argless_boundary(self):
        @boundary
        def argless():
            yield False  # receive args
            yield False  # allow first yield
            yield True

        # the following simulates:
        #   @argless
        #   def fibonacci():
        #      ....
        fib2 = argless(fibonacci)
        self.assertEqual(fib2(), 1)


class TestHigherLevelPreds(unittest.TestCase):
    def test_close_is_always_called(self):
        @boundary
        def bailout():
            yield
            try:
                yield False
                yield True
            except GeneratorExit:
                pass
            else:
                raise AssertionError("close() must have been called")

        fibnone = whenall(whenany(bailout))(fibonacci)
        self.assertEqual(fibnone(), 1)

    def test_whenall_with_invalid(self):
        from xotl.tools.bound import times

        @boundary
        def invalid():
            yield

        fibinv = whenall(invalid, times(10))(fibonacci)
        with self.assertRaises(RuntimeError):
            fibinv()

    def test_whenall_with_invalid_befored_terminated(self):
        from xotl.tools.bound import times

        @boundary
        def invalid():
            yield
            yield False

        fibinv = whenall(invalid, times(10))(fibonacci)
        with self.assertRaises(RuntimeError):
            fibinv()


class TestBoundedUnnamedPredicates(unittest.TestCase):
    def test_atmost_unnamed(self):
        from xotl.tools.bound import times

        fib8 = times(8)(fibonacci)
        # Fibonacci numbers are yielded:
        # 1 1 2 3 5 8 13 21
        self.assertEqual(fib8(), 21)

    def test_invalid_unnamed(self):
        @boundary
        def invalid():
            return 1

        @invalid
        def foobar():
            while True:
                yield

        with self.assertRaises(TypeError):
            foobar()

        @boundary
        def invalid_init():
            yield

        @invalid_init
        def foobar2():
            while True:
                yield

        with self.assertRaises(RuntimeError):
            foobar2()


class TestBoundedPredicates(unittest.TestCase):
    def test_invalid_predicate_early_at_init(self):
        @boundary
        def invalid():
            yield

        @invalid
        def foobar():
            pass

        with self.assertRaises(RuntimeError):
            foobar()

    def test_invalid_predicate_early_at_cycle(self):
        @boundary
        def invalid():
            yield
            yield False  # i.e never signal True

        @invalid
        def foobar():
            passes, atmost = 0, 10
            while passes < atmost:
                yield passes
                passes += 1
            raise AssertionError("Invalid reach point a GeneratorExit was expected.")

        with self.assertRaises(RuntimeError):
            foobar()


class TestMisc(unittest.TestCase):
    def test_args_are_passed(self):
        @boundary
        def pred():
            args, kwargs = yield
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {})
            yield True

        @pred
        def foobar(*args, **kwargs):
            while True:
                yield 1

        foobar(1, 2)

    def test_whens_receives_args(self):
        from xotl.tools.bound import whenall, whenany

        self.assertTrue(whenall.receive_args)
        self.assertTrue(whenany.receive_args)

    def test_args_are_passed_to_all(self):
        from xotl.tools.bound import whenall, whenany

        @boundary
        def pred():
            args, kwargs = yield
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {"egg": "ham"})
            yield True

        @whenall(pred, pred())
        def foobar(*args, **kwargs):
            while True:
                yield 1

        foobar(1, 2, egg="ham")

        @whenany(pred(), pred)
        def foobar(*args, **kwargs):
            while True:
                yield 1

        foobar(1, 2, egg="ham")

    def test_needs_args(self):
        from xotl.tools.bound import times

        @whenall(times)
        def foobar():
            yield

        with self.assertRaises(TypeError):
            foobar()  # times is not initialized

    def test_plain_function(self):
        def pred():
            args, kwargs = yield
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {"egg": "ham"})
            yield True

        @whenall(pred)
        def foobar(*args, **kwargs):
            while True:
                yield 1

        foobar(1, 2, egg="ham")

    def test_generators(self):
        def pred():
            args, kwargs = yield
            self.assertEqual(args, (1, 2))
            self.assertEqual(kwargs, {"egg": "ham"})
            yield True

        @whenall(pred())  # a generator!!
        def foobar(*args, **kwargs):
            while True:
                yield 1

        foobar(1, 2, egg="ham")

    def test_plain_generator(self):
        from xotl.tools.bound import times

        fibseq = fibonacci()
        limited = times(5)(fibseq)
        self.assertEqual(limited(), 5)


class TestTerminationCases(unittest.TestCase):
    def test_termination_when_exception(self):
        def magic_number(a):
            if a < 1:
                yield 0
            else:
                prev = next(magic_number(a - 1))
                yield a + prev

        @boundary
        def forever():
            try:
                while True:
                    yield False
            except GeneratorExit:
                pass  # ok
            else:
                raise AssertionError("close() not called to boundary")

        bounded = forever()(magic_number)
        bounded(0)  # No exception
        with self.assertRaises(TypeError):
            bounded("invalid")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from xotl.tools.tasking.safe import SafeData


def test_safe():
    from xotl.tools.future.threading import async_call
    from time import sleep

    data = {}
    qd = SafeData(data, timeout=5.0)

    def inner(name, start, end):
        step = 1 if start < end else -1
        for i in range(start, end, step):
            with qd as d:
                d[i] = d.get(i, 0) + 1
            sleep(0.001 * i)
        with qd as d:
            d[name] = True

    one, two = "one", "two"
    async_call(inner, args=[one, 1, 6])
    async_call(inner, args=[two, 8, 3])
    finish = {one: False, two: False}
    while not (finish[one] and finish[two]):
        with qd as d:
            for k in (one, two):
                if d.get(k):
                    finish[k] = True
        sleep(0.001)
    aux = {i: 1 for i in range(1, 6)}
    for i in range(8, 3, -1):
        aux[i] = aux.get(i, 0) + 1
    aux[one] = aux[two] = True
    assert data == aux



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest

from contextlib import contextmanager
from datetime import datetime, timedelta

from xotl.tools.future.functools import lru_cache


@lru_cache(3)
def fib(n):
    print(n)
    if n <= 1:
        return 1
    else:
        # It seems that there's a difference in the execution path for `return
        # fib(n-2) + fib(n-1)` between Python 2.7 and Python 3.2, so let's make
        # more explicit the order we'd like so the test is more reliable.
        a = fib(n - 1)
        b = fib(n - 2)
        return a + b


def takes_no_more_than(duration, msg=None):
    if not msg:
        msg = "It took longer than {s} seconds".format(s=duration)

    @contextmanager
    def inner():
        start = datetime.now()
        yield
        end = datetime.now()
        max_duration = timedelta(seconds=duration)
        if (end - start) > max_duration:
            raise AssertionError(msg)

    return inner()


def test_lrucache():
    # Without caching fib(120) would take ages.  On a 2.20GHz laptop with
    # caching this takes less than 1 sec, so let's test that it will respond in
    # no more than 3 min to allow very slow machines testing this code.
    fib.cache_clear()
    with takes_no_more_than(90):
        assert fib(120) == 8670007398507948658051921


def test_lrucache_stats():
    pass


from xotl.tools.fp.tools import compose, identity


class TestCompose(unittest.TestCase):
    def test_needs_at_least_an_argument(self):
        self.assertIs(compose(), identity)

    def test_single_argument_is_identitical(self):
        def anything():
            pass

        self.assertIs(anything, compose(anything))

    def test_only_callables(self):
        with self.assertRaises(TypeError):
            compose(1)

    def test_simple_case(self):
        incr = lambda x: x + 1
        add_3 = compose(incr, incr, incr)
        self.assertEqual(3, add_3(0))

    def test_with_pow(self):
        from xotl.tools.future.functools import power

        incr = lambda x: x + 1
        add_1 = power(incr, 1)
        self.assertIs(incr, add_1)
        add_3 = power(incr, 3)
        self.assertEqual(3, add_3(0))


def test_lwraps():
    from xotl.tools.future.functools import lwraps

    class foobar:
        @lwraps("method-one", one=True)
        def one(self):
            return type(self).__name__

        @lwraps("method-two", two=True)
        @classmethod
        def two(cls):
            return cls.__name__

        @lwraps("method-three", three=True)
        @staticmethod
        def three():
            return "foobar"

    @lwraps("function-four", four=True, one=False)
    def four(*args):
        return [(arg.__name__, arg()) for arg in args]

    f = foobar()
    names = ("method-one", "method-two", "method-three", "function-four")

    assert four.__name__ == names[3]
    assert foobar.one.__name__ == names[0]
    assert foobar.two.__name__ == names[1]
    assert foobar.three.__name__ == names[2]
    assert f.one.__name__ == names[0]
    assert f.two.__name__ == names[1]
    assert f.three.__name__ == names[2]

    assert four.four
    assert not four.one
    assert f.one.one
    assert f.two.two
    assert f.three.three

    for i, (a, b) in enumerate(four(f.one, f.two, f.three)):
        assert a == names[i]
        assert b == "foobar"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Python Software Foundation
# All rights reserved.
#
# Most of the contents of this file were extracted from the source code of
# CPython 3.6.
#

"""Utilities for `with-statement contexts <343>`:pep:.

This module re-export all symbols from the standard library, with the
exception of the function `nested`.

.. versionadded:: 1.9.5

"""

import contextlib as _stdlib  # noqa
from contextlib import *  # noqa

__all__ = list(getattr(_stdlib, "__all__", []))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest
import random

from xotl.tools.bases import B32, B64, B64symbolic


def build_many_tests():
    def test_many_random_convertions(self):
        subjects = [B32, B64, B64symbolic]
        for _ in range(5):
            testcase = random.randrange(2 ** 64, 2 ** 128)
            subject = random.choice(subjects)
            assert testcase == subject.basetoint(subject.inttobase(testcase))

    return {
        "test_many_random_convertions_%d" % i: test_many_random_convertions
        for i in range(10)
    }


_TestManyConvertions = type(str("TestManyConvertions"), (object,), build_many_tests())


class TestManyConvertions(unittest.TestCase, _TestManyConvertions):
    def test_regression_56107046767814579580835126010035242071(self):
        number = 56107046767814579580835126010035242071
        b64 = B64symbolic
        assert number == b64.basetoint(b64.inttobase(number))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from hypothesis import given, example
from hypothesis.strategies import text


def test_slugify():
    from xotl.tools.string import slugify

    value = "  Á.e i  Ó  u  "
    options = dict(encoding="utf-8")
    assert slugify(value, **options) == "a-e-i-o-u"
    assert slugify(value, ".", invalid_chars="AU", **options) == "e.i.o"
    assert slugify(value, valid_chars=".", **options) == "a.e-i-o-u"
    assert slugify("_x", "_") == "_x"
    assert slugify("-x", "_") == "x"
    assert slugify("-x-y-", "_") == "x_y"
    assert slugify(None) == "none"
    assert slugify(1 == 1) == "true"
    assert slugify(1.0) == "1-0"
    assert slugify(135) == "135"
    assert slugify(123456, "", invalid_chars="52") == "1346"
    assert slugify("_x", "_") == "_x"


# FIXME: Dont filter; `slugify` should consider this.
valid_replacements = text().filter(lambda x: "\\" not in x)


@given(s=text(), invalid_chars=text(), replacement=valid_replacements)
@example(s="0/0", invalid_chars="-", replacement="-")
def test_slugify_hypothesis(s, invalid_chars, replacement):
    # TODO: (s='0:0', invalid_chars='z', replacement='ź')
    from xotl.tools.string import slugify
    from xotl.tools.string import force_ascii

    assert " " not in slugify(s), "Slugs do not contain spaces"

    assert " " in slugify(
        s + " ", valid_chars=" "
    ), "Slugs do contain spaces if explicitly allowed"

    replacement = force_ascii(replacement).lower()
    invalid_chars = force_ascii(invalid_chars).lower()
    assert all(
        c not in slugify(s, replacement, invalid_chars=c)
        for c in invalid_chars
        if c not in replacement
    ), "Slugs dont contain invalid chars"


@given(s=text(), p=text())
def test_cutting_is_inverse_to_adding(s, p):
    from xotl.tools.string import cut_prefix, cut_suffix

    assert cut_prefix(p + s, p) == s
    assert cut_suffix(s + p, p) == s
    assert cut_suffix(s, "") == s
    assert cut_prefix(s, "") == s


@given(s=text(), p=text())
def test_cutting_is_stable(s, p):
    from xotl.tools.string import cut_prefix, cut_suffix

    if not s.startswith(p):
        assert cut_prefix(s, p) == s == cut_prefix(cut_prefix(s, p), p)
    if not s.endswith(p):
        assert cut_suffix(s, p) == s == cut_suffix(cut_suffix(s, p), p)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Tools for manage Python keywords as names.

Reserved Python keywords can't be used as attribute names, so this module
functions use the convention of rename the name using an underscore as
suffix when a reserved keyword is used as name.

"""


def suffix_kwd(name):
    """Add an underscore suffix if name if a Python keyword."""
    from keyword import iskeyword

    return "{}_".format(name) if iskeyword(name) else name


def org_kwd(name):
    """Remove the underscore suffix if name starts with a Python keyword."""
    from keyword import iskeyword

    if name.endswith("_"):
        res = name[:-1]
        return res if iskeyword(res) else name
    else:
        return name


def getkwd(obj, name, default=None):
    """Like `getattr` but taking into account Python keywords."""
    return getattr(obj, suffix_kwd(name), default)


def setkwd(obj, name, value):
    """Like `setattr` but taking into account Python keywords."""
    setattr(obj, suffix_kwd(name), value)


def delkwd(obj, name):
    """Like `delattr` but taking into account Python keywords."""
    delattr(obj, suffix_kwd(name))


def kwd_getter(obj):
    """partial(getkwd, obj)"""
    from functools import partial

    return partial(getkwd, obj)


def kwd_setter(obj):
    """partial(setkwd, obj)"""
    from functools import partial

    return partial(setkwd, obj)


def kwd_deleter(obj):
    """partial(delkwd, obj)"""
    from functools import partial

    return partial(delkwd, obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import pytest

from xotl.tools.future.datetime import date, datetime, timedelta
from xotl.tools.future.datetime import daterange
from xotl.tools.future.datetime import TimeSpan, EmptyTimeSpan, DateTimeSpan
from xotl.tools.testing.datetime import timespans, datetimespans

import hypothesis
from hypothesis import strategies, given, settings


dates = strategies.dates
maybe_date = dates() | strategies.none()
maybe_datetimes = strategies.datetimes() | strategies.none()


def test_datetime_imports():
    from xotl.tools.future import datetime
    from xotl.tools.future import datetime as dt

    assert datetime is dt

    from xotl.tools.future.datetime import TimeSpan
    from xotl.tools.future.datetime import TimeSpan as TS

    assert TS is TimeSpan


def test_daterange_stop_only():
    result = list(daterange(date(1978, 10, 21)))
    assert result[0] == date(1978, 10, 1)
    assert result[-1] == date(1978, 10, 20)


def test_daterange_empty():
    assert [] == list(daterange(date(1978, 10, 21), -2))
    assert [] == list(daterange(date(1978, 10, 21), date(1978, 10, 10)))
    assert [] == list(daterange(date(1978, 10, 10), date(1978, 10, 20), -1))


def test_daterange_going_back_in_time():
    result = list(daterange(date(1978, 10, 21), -2, -1))
    assert result[0] == date(1978, 10, 21)
    assert result[-1] == date(1978, 10, 20)


def test_daterange_invalid_int_stop():
    with pytest.raises(TypeError):
        daterange(10)


def test_daterange_invalid_step():
    with pytest.raises(ValueError):
        daterange(None, date(1978, 10, 21), 0)


@given(timespans(), timespans())
@hypothesis.example(ts1=TimeSpan(), ts2=timespans().example())
def test_intersection_commutable(ts1, ts2):
    # Commutable
    assert ts2 * ts1 == (ts1 & ts2)


@given(timespans(), timespans())
@hypothesis.example(ts1=TimeSpan(), ts2=timespans().example())
def test_intersection_containment(ts1, ts2):
    overlap = ts1 * ts2
    if overlap is not None:
        # The intersection must always be totally covered by both ts1 and ts2,
        # unless ts1 and ts2 don't intersect
        assert (overlap & ts1) == overlap
        assert (overlap <= ts1) is True
        assert (overlap <= ts2) is True


@given(timespans(), timespans())
@hypothesis.example(ts1=TimeSpan(), ts2=timespans().example())
def test_comparision(ts1, ts2):
    if ts1 <= ts2 <= ts1:
        assert ts1 == ts2
    if ts1 == ts2:
        assert ts1 <= ts2 <= ts1

    # Single day intersection and equality test
    if ts1.start_date:
        assert ts1 * ts1.start_date == ts1.start_date
        assert ts1.start_date * ts1 == ts1.start_date
        assert ts1.start_date in ts1


@given(timespans(), datetimespans())
def test_interaction_timespan_with_datetimespans(ts, dts):
    assert isinstance(dts & ts, (DateTimeSpan, type(EmptyTimeSpan)))
    assert isinstance(ts & dts, (DateTimeSpan, type(EmptyTimeSpan)))

    assert DateTimeSpan.from_timespan(ts) == ts
    assert ts == DateTimeSpan.from_timespan(ts)

    ts = TimeSpan("2018-01-01", "2018-01-01")
    dts = DateTimeSpan("2018-01-01 00:00:01", "2018-01-01 00:00:02")

    assert ts != dts
    assert dts != ts


@given(timespans(), timespans(), dates())
def test_general_cmp_properties(ts1, ts2, date):
    assert bool(ts1 <= ts2) == bool(ts2 >= ts1)
    # In Python 2, dates have a __le__ that does no compare to timespans.
    assert bool(TimeSpan.from_date(date) <= ts2) == bool(ts2 >= date)

    overlap = ts1 & ts2
    if not overlap:
        # Disjoint sets are not orderable...
        assert not (ts1 <= ts2) and not (ts2 <= ts1)


@given(timespans(unbounds="future"))
def test_outside_date(ts):
    from datetime import timedelta

    assert ts.start_date
    outsider = ts.start_date - timedelta(1)
    assert outsider not in ts


@given(
    datetimespans(
        dates=strategies.datetimes(min_value=datetime(1, 1, 2)), unbounds="future"
    )
)
def test_outside_datetime(dts):
    from datetime import timedelta

    assert dts.start_date
    outsider = dts.start_date - timedelta(1)
    assert outsider not in dts


@given(timespans() | datetimespans())
def test_empty_timespan(ts):
    assert ts >= EmptyTimeSpan <= ts, "Empty is a subset of any TS"

    assert (
        EmptyTimeSpan <= EmptyTimeSpan >= EmptyTimeSpan
    ), "Empty is a subset of itself"

    assert not EmptyTimeSpan, "Empty is considered False"

    assert not (ts <= EmptyTimeSpan), "Empty is not a superset of any TS"

    type(EmptyTimeSpan)() is EmptyTimeSpan

    assert EmptyTimeSpan & ts == EmptyTimeSpan * ts == EmptyTimeSpan
    assert EmptyTimeSpan | ts == EmptyTimeSpan + ts == ts


@given(timespans(unbounds="none"), timespans())
def test_failure_of_triple_intersection(ts1, ts2):
    from datetime import timedelta

    ts0 = TimeSpan.from_date(ts1.start_date - timedelta(1))
    assert not (ts0 & ts1 & ts2)


@given(strategies.dates())
def test_dates_are_representable(value):
    from xotl.tools.future.datetime import date

    class mydate(date):
        pass

    value = mydate(value.year, value.month, value.day)
    assert value.strftime("%Y-%m-%d")


@given(timespans(unbounds="none"))
def test_timespans_are_representable(value):
    assert repr(value)


@given(timespans(unbounds="none"))
def test_generate_valid_timespans(ts):
    assert ts.valid


@given(timespans(unbounds="none"))
def test_ts_returns_dates_not_subtypes(ts):
    from datetime import date

    assert type(ts.start_date) is date
    assert type(ts.end_date) is date


@given(timespans(unbounds="none"), strategies.dates())
def test_operate_with_timespans(ts, d):
    assert ts.start_date - d is not None
    assert d - ts.start_date is not None


@given(timespans(unbounds="none"), timespans(unbounds="none"))
def test_definition_of_overlaps(ts1, ts2):
    assert ts1.overlaps(ts2) == bool(ts1 & ts2)


@given(timespans(unbounds="none"))
def test_duplication_of_timespans(ts1):
    ts2 = TimeSpan(ts1.start_date, ts1.end_date)
    assert {ts1, ts2} == {ts1}, "ts1 and ts2 are equal but different!"


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_reversed(ts1, delta):
    try:
        assert (ts1 << delta) == (ts1 >> -delta)
    except OverflowError:
        pass


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_keeps_unbounds(ts1, delta):
    try:
        assert ts1.unbound == (ts1 << delta).unbound
        assert ts1.future_unbound == (ts1 << delta).future_unbound
        assert ts1.past_unbound == (ts1 << delta).past_unbound
    except OverflowError:
        pass


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_doubled(ts1, delta):
    try:
        assert ((ts1 << delta) << delta) == (ts1 << (2 * delta))
    except OverflowError:
        pass


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_backandforth(ts1, delta):
    try:
        assert ts1 == ((ts1 << delta) >> delta) == (ts1 << 0) == (ts1 >> 0)
    except OverflowError:
        pass


@given(timespans(unbounds="none"), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_dates(ts1, delta):
    try:
        res = ts1 << delta
    except OverflowError:
        # Ignore if the date it's being displaced to non-supported date,
        # that's for the client to deal with
        pass
    else:
        assert (res.start_date - ts1.start_date).days == -delta
        assert (res.end_date - ts1.end_date).days == -delta
    try:
        res = ts1 >> delta
    except OverflowError:
        # Ignore if the date it's being displaced to non-supported date,
        # that's for the client to deal with
        pass
    else:
        assert (res.start_date - ts1.start_date).days == delta
        assert (res.end_date - ts1.end_date).days == delta


@given(
    timespans(unbounds="none") | datetimespans(unbounds="none"),
    strategies.integers(min_value=-1000, max_value=1000),
)
def test_timespans_displacement_keeps_the_len(ts1, delta):
    try:
        res = ts1 << delta
    except OverflowError:
        pass
    else:
        assert len(ts1) == len(res)


@given(timespans() | datetimespans())
def test_timespans_are_pickable(ts):
    import pickle

    for proto in range(1, pickle.HIGHEST_PROTOCOL + 1):
        assert ts == pickle.loads(pickle.dumps(ts, proto))


def test_empty_timespan_is_pickable():
    import pickle

    for proto in range(1, pickle.HIGHEST_PROTOCOL + 1):
        assert EmptyTimeSpan is pickle.loads(pickle.dumps(EmptyTimeSpan, proto))


@given(strategies.datetimes(), strategies.datetimes())
def test_timespan_with_datetimes(d1, d2):
    from datetime import datetime as dt, date as d

    ts = TimeSpan(d1, d2)
    assert not isinstance(ts.start_date, dt)
    assert isinstance(ts.start_date, d)
    assert not isinstance(ts.end_date, dt)
    assert isinstance(ts.end_date, d)


@given(datetimespans())
def test_datetimespans_ts_fields(dts):
    if dts.start_datetime is None:
        assert dts.start_date is None
    else:
        assert dts.start_date == dts.start_datetime.date()
    if dts.end_datetime is None:
        assert dts.end_date is None
    else:
        assert dts.end_date == dts.end_datetime.date()


@given(maybe_date, datetimespans())
def test_setting_start_date(d, dts):
    dts.start_date = d
    if d is not None:
        assert dts.start_datetime.date() == d
        assert dts.start_datetime == datetime(d.year, d.month, d.day)
    else:
        assert dts.start_datetime is None


@given(maybe_date, datetimespans())
def test_setting_end_date(d, dts):
    dts.end_date = d
    if d is not None:
        assert dts.end_datetime.date() == d
        assert dts.end_datetime == datetime(d.year, d.month, d.day, 23, 59, 59)
    else:
        assert dts.end_datetime is None


@given(
    strategies.datetimes(
        min_value=datetime(1970, 1, 1), max_value=datetime(5000, 12, 31)
    ),
    strategies.integers(min_value=1, max_value=200),
)
def test_timespan_diff(start_date, delta):
    days = timedelta(days=delta)
    big = TimeSpan(start_date, start_date + 3 * days)
    x, y = big.diff(big)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan

    x, y = big.diff(EmptyTimeSpan)
    assert x == big and y is EmptyTimeSpan

    outsider = big >> 4 * days
    assert not big & outsider
    x, y = big.diff(outsider)
    assert x == big and y is EmptyTimeSpan
    x, y = outsider.diff(big)
    assert x == outsider and y is EmptyTimeSpan

    atstart = TimeSpan(start_date, start_date + days)
    assert atstart < big
    x, y = big.diff(atstart)
    assert x is EmptyTimeSpan
    assert y
    assert add_timespans(atstart, y) == big

    beforestart = TimeSpan(start_date - days, start_date + days)
    x, y = big.diff(beforestart)
    assert x is EmptyTimeSpan
    assert y
    assert add_timespans(beforestart & big, y) == big

    atend = TimeSpan(start_date + 2 * days, start_date + 3 * days)
    assert atend < big
    x, y = big.diff(atend)
    assert y is EmptyTimeSpan
    assert x
    assert add_timespans(x, atend) == big

    afterend = TimeSpan(start_date + 2 * days, start_date + 4 * days)
    x, y = big.diff(afterend)
    assert y is EmptyTimeSpan
    assert x
    assert add_timespans(x, big & afterend) == big

    middle = TimeSpan(start_date + days, start_date + 2 * days)
    assert middle < big
    x, y = big.diff(middle)
    assert x and y
    assert add_timespans(add_timespans(x, middle), y) == big

    bigger = TimeSpan(start_date - days, start_date + 4 * days)
    x, y = big.diff(bigger)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan


@given(
    strategies.datetimes(
        min_value=datetime(1970, 1, 1), max_value=datetime(5000, 12, 31)
    ),
    strategies.integers(min_value=1, max_value=10000),
)
@settings(deadline=None)
def test_datetimespan_diff(start_date, delta):
    secs = timedelta(seconds=delta)
    big = DateTimeSpan(start_date, start_date + 3 * secs)
    assert big.start_datetime == start_date
    assert big.end_datetime == start_date + 3 * secs
    x, y = big.diff(big)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan

    x, y = big.diff(EmptyTimeSpan)
    assert x == big and y is EmptyTimeSpan

    outsider = big >> 4 * secs
    assert not big & outsider
    x, y = big.diff(outsider)
    assert x == big and y is EmptyTimeSpan
    x, y = outsider.diff(big)
    assert x == outsider and y is EmptyTimeSpan

    atstart = DateTimeSpan(start_date, start_date + secs)
    assert atstart < big
    x, y = big.diff(atstart)
    assert x is EmptyTimeSpan
    assert y
    assert add_dtspans(atstart, y) == big

    beforestart = DateTimeSpan(start_date - secs, start_date + secs)
    x, y = big.diff(beforestart)
    assert x is EmptyTimeSpan
    assert y
    assert add_dtspans(beforestart & big, y) == big

    atend = DateTimeSpan(start_date + 2 * secs, start_date + 3 * secs)
    assert atend < big
    x, y = big.diff(atend)
    assert y is EmptyTimeSpan
    assert x
    assert add_dtspans(x, atend) == big

    afterend = DateTimeSpan(start_date + 2 * secs, start_date + 4 * secs)
    x, y = big.diff(afterend)
    assert y is EmptyTimeSpan
    assert x
    assert add_dtspans(x, big & afterend) == big

    middle = DateTimeSpan(start_date + secs, start_date + 2 * secs)
    assert middle < big
    x, y = big.diff(middle)
    assert x and y
    assert add_dtspans(add_dtspans(x, middle), y) == big

    bigger = DateTimeSpan(start_date - secs, start_date + 4 * secs)
    x, y = big.diff(bigger)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan


@given(datetimespans())
def test_datetimespan_repr(ts):
    assert eval(repr(ts)) == ts
    assert DateTimeSpan("2018-01-01") == DateTimeSpan(datetime(2018, 1, 1))


def add_timespans(x, y):
    if x.end_date == y.start_date - timedelta(1):
        return TimeSpan(x.start_date, y.end_date)
    else:
        raise ValueError


def add_dtspans(x, y):
    if x.end_datetime == y.start_datetime - timedelta(seconds=1):
        return DateTimeSpan(x.start_datetime, y.end_datetime)
    else:
        raise ValueError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import pytest
from hypothesis import given, strategies as s

from xotl.tools.dim.meta import Signature, Quantity, Dimension, Scalar, UNIT


# A reasonable exponent.  We won't be dealing with universes of 100s
# dimensions.
exponents = s.integers(min_value=1, max_value=10)
signatures = s.text(alphabet="abcdefghijok", min_size=0, max_size=10)


def test_usage():
    @Dimension.new
    class L:
        metre = UNIT
        kilometre = 1000 * metre

    @Dimension.new
    class T:
        second = UNIT

    m, km = L.metre, L.kilometre
    s = T.second

    assert isinstance(m, L)
    assert not isinstance(1, L)
    assert isinstance(s, T)

    assert -km == -1 * km == -(1000 * m)
    assert +km == km

    Speed = L / T

    assert isinstance(m / s, Speed)
    assert not isinstance(m, T)

    assert hasattr(L / T, "metre_per_second")

    assert 10 * km == 10000 * m
    assert m < km

    with pytest.raises(TypeError):
        10 + 10 * km

    with pytest.raises(TypeError):
        m + s

    assert Speed.metre_per_second == m / s

    Acceleration = L / (T * T)
    assert hasattr(Acceleration, "metre_per_second_squared")
    # however:
    Not_Acceleration = L / T * T
    assert hasattr(Not_Acceleration, "metre_per_second_second")
    assert L == Not_Acceleration, "It is the same as Length"


def test_effort():
    @Dimension.new
    class Workforce:
        men = UNIT

    @Dimension.new
    class Time:
        second = UNIT

    class Effort(Workforce * Time):
        # Since the canonical unit of a composed quantity type is built from
        # the canonical units of the operands, but the true "canonical type"
        # of effort is usually men-hour we re-introduce it.
        men_hour = 60

    assert Effort.men_hour > Effort.men_second


def test_scalar_downgrade():
    from xotl.tools.dim.base import L

    km = L.km
    assert not isinstance(km / km, Quantity)
    assert km / km == 1
    assert not isinstance(km // km, Quantity)
    assert km // km == 1

    assert not isinstance(1 / km * km, Quantity)
    assert float(1 / km * km)


def test_natural_downgrade():
    from xotl.tools.dim.base import L

    km, cm = L.km, L.cm
    assert float(km) == 1000
    assert int(cm) == 0


def test_decimals():
    import decimal
    from xotl.tools.dim.base import m

    with pytest.raises(TypeError):
        third = decimal.Decimal("0.33") * m
        assert third < m


def test_signatures():
    distance = Signature("m")
    time = Signature("s")
    freq = 1 / time
    speed = distance / time
    acceleration = speed / time
    assert acceleration == distance / (time * Signature("s"))
    assert speed == distance * freq
    assert speed ** 3 == speed * speed * speed
    assert speed ** 0 == Signature() == speed / speed
    assert speed ** -3 == 1 / (speed ** 3)


@given(signatures, signatures)
def test_signatures_are_always_simplified(top, bottom):
    s = Signature(top, bottom)
    assert all(t not in s.bottom for t in s.top)
    assert all(b not in s.top for b in s.bottom)
    r = Signature(bottom, top)
    assert s.top == r.bottom and s.bottom == r.top

    s2 = Signature(top + bottom, bottom)
    assert s2 == Signature(top, None)
    s3 = Signature(top, top + bottom)
    assert s3 == Signature(None, bottom)


def test_quantity_math():
    metre = m = Quantity(1, Signature("m"))
    second = Quantity(1, Signature("s"))

    with pytest.raises(TypeError):
        metre < second

    assert metre < 2 * metre < (metre + 2 * metre)
    assert metre * metre == Quantity(1, Signature("mm"))
    assert metre / second == Quantity(1, Signature("m", "s"))
    assert metre * metre * metre == metre ** 3
    assert 1 / (metre * metre * metre) == metre ** -3

    assert 1000 * m % 3 == 1 * m
    assert 5 % (2 * m) == 1 / m
    assert 5 / (2 * m) == 2.5 / m

    with pytest.raises(TypeError):
        5 ** (2 * m)


def test_quantity_type_definitions():
    from xotl.tools.dim.base import Length, Time

    assert isinstance(Length, Dimension)
    assert isinstance(Time, Dimension)
    assert isinstance(Length / Time, Dimension)
    assert isinstance(Length ** 2, Dimension)
    assert Length * Length == Length ** 2

    assert Time / Time == Scalar
    assert Time / Time * Time == Time

    with pytest.raises(TypeError):
        Length + Time

    with pytest.raises(TypeError):
        Length - Time

    assert Length ** 1 is Length

    with pytest.raises(TypeError):
        Length ** 1.2

    assert Length ** 0 == Scalar
    assert Length ** -1 == 1 / Length

    with pytest.raises(TypeError):
        2 / Length

    with pytest.raises(TypeError):
        2 * Length


@given(exponents, exponents)
def test_general_power_rules(n, m):
    from xotl.tools.dim.base import L

    assert L ** n / L ** m == L ** (n - m)


@given(s.floats(allow_nan=False) | s.integers())
def test_any_magnitude(m):
    from xotl.tools.dim.base import L

    assert float(m * L.metre) == float(m)


@given(s.floats(allow_nan=False, allow_infinity=False) | s.integers())
def test_any_magnitude_noinf(m):
    from xotl.tools.dim.base import L
    from math import ceil, floor

    Int = int
    q = m * L.metre
    for f in (Int, float, abs, round, ceil, floor):
        assert f(q) == f(m)


def test_currencies():
    from xotl.tools.dim.currencies import Rate, Valuation, currency

    dollar = USD = currency("USD")
    euro = EUR = currency("EUR")
    rate = 1.19196 * USD / EUR

    assert isinstance(dollar, Valuation)
    assert isinstance(rate, Rate)

    # Even 0 dollars are a valuation
    assert isinstance(dollar - dollar, Valuation)

    # But 1 is not a value nor a rate
    assert not isinstance(dollar / dollar, Valuation)
    assert not isinstance(dollar / dollar, Rate)

    assert currency("a") is currency("A")

    with pytest.raises(TypeError):
        dollar + euro


def test_undistinguishable_definitions():
    from xotl.tools.dim.base import L

    @Dimension.new
    class Length:
        metre = UNIT

    assert L.metre == Length.metre

    @Dimension.new
    class Length:
        km = UNIT

    with pytest.raises(TypeError):
        assert L.metre != Length.km


def test_bug_30():
    from xotl.tools.dim.meta import Dimension, UNIT

    @Dimension.new
    class L:
        m = UNIT

    with pytest.raises(TypeError):

        class LL(L):
            mm = UNIT


def test_custom_quantity():
    class NewQuantity(Quantity):
        pass

    class NewDim(Dimension):
        Quantity = NewQuantity

    @NewDim.new
    class Length:
        m = UNIT

    m = Length.m
    Area = Length ** 2

    assert isinstance(m, NewQuantity)
    assert isinstance(1000 * m, NewQuantity)
    assert isinstance(Area._unit_, NewQuantity)

    @Dimension.new(Quantity=NewQuantity)
    class Time:
        s = UNIT

    Freq = 1 / Time
    assert isinstance(Time.s, NewQuantity)
    assert isinstance(Freq._unit_, NewQuantity)

    @NewDim.new(Quantity=Quantity)
    class Time:
        s = UNIT

    Freq = 1 / Time
    assert not isinstance(Time.s, NewQuantity)
    assert isinstance(Time.s, Quantity)
    assert not isinstance(Freq._unit_, NewQuantity)
    assert isinstance(Freq._unit_, Quantity)



from typing import Any, Callable, Dict, Type, Union, TypeVar, overload, Tuple
from typing_extensions import Literal

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
Z = TypeVar("Z")
X = TypeVar("X")
Y = TypeVar("Y")

def identity(arg: C) -> C: ...
@overload
def compose(
    f: Callable[[A], B],
    g: Callable[[C], A],
) -> Callable[[C], B]: ...
@overload
def compose(
    z: Callable[[B], Z],
    f: Callable[[A], B],
    g: Callable[[C], A],
) -> Callable[[C], Z]: ...
@overload
def compose(
    x: Callable[[Z], X],
    z: Callable[[B], Z],
    f: Callable[[A], B],
    g: Callable[[C], A],
) -> Callable[[C], X]: ...
@overload
def compose(
    y: Callable[[X], Y],
    x: Callable[[Z], X],
    z: Callable[[B], Z],
    f: Callable[[A], B],
    g: Callable[[C], A],
) -> Callable[[C], Y]: ...
@overload
def compose(*fn: Callable[[A], A]) -> Callable[[A], A]: ...
def constant(A) -> Callable[..., A]: ...
@overload
def fst(pair: Tuple[A, B]) -> A: ...
@overload
def fst(pair: Tuple[A, B], strict: Literal[True]) -> A: ...
@overload
def fst(pair: Tuple[A, ...], strict: Literal[False]) -> A: ...
@overload
def snd(pair: Tuple[A, B]) -> B: ...
@overload
def snd(pair: Tuple[A, B], strict: Literal[True]) -> B: ...
@overload
def snd(pair: Tuple[B, ...], strict: Literal[False]) -> B: ...



from typing import Callable, Iterable, TypeVar, overload

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
@overload
def kleisli_compose(
    g: Callable[[B], Iterable[C]],
    f: Callable[[A], Iterable[B]],
) -> Callable[[A], Iterable[C]]: ...
@overload
def kleisli_compose(*fs: Callable[[A], Iterable[A]]) -> Callable[[A], Iterable[A]]: ...
@overload
def kleisli_compose_foldl(
    g: Callable[[B], Iterable[C]],
    f: Callable[[A], Iterable[B]],
) -> Callable[[A], Iterable[C]]: ...
@overload
def kleisli_compose_foldl(
    *fs: Callable[[A], Iterable[A]]
) -> Callable[[A], Iterable[A]]: ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Codec registry, base classes and tools.

In this module, some additions for `codecs` standard module.

"""


from codecs import *  # noqa
from codecs import __all__  # noqa

__all__ = list(__all__)


def force_encoding(encoding=None):
    """Validates an encoding value.

    If `encoding` is None use `locale.getdefaultlocale`:func:.  If that is
    also none, return 'UTF-8'.

    .. versionadded:: 1.2.0

    .. versionchanged:: 1.8.0 migrated to 'future.codecs'

    .. versionchanged:: 1.8.7 Stop using `locale.getpreferrededencoding`:func:
       and improve documentation.

    """
    # TODO: This mechanism is tricky, we must find out how to unroll the mess
    # involving the concept of which encoding to use by default:
    #
    # - locale.getlocale(): In Python 2 returns ``(None, None)``, but in
    #   Python 3 ``('en_US', 'UTF-8')``.
    #
    # - locale.getpreferredencoding(): all versions returns ``'UTF-8'``.
    #
    # - sys.getdefaultencoding(): In Python 2 returns ``'ascii'``, but in
    #   Python 3 ``'utf-8'``.  The same in Mac-OS. The related code was
    #   commented because these differences.
    #
    # All these considerations where also proved in Mac-OS.
    import locale

    return encoding or locale.getdefaultlocale()[1] or "UTF-8"


def safe_decode(s, encoding=None):
    """Similar to bytes `decode` method returning unicode.

    Decodes `s` using the given `encoding`, or determining one from the system.

    Returning type depend on python version; if 2.x is `unicode` if 3.x `str`.

    .. versionadded:: 1.1.3
    .. versionchanged:: 1.8.0 migrated to 'future.codecs'

    """
    if isinstance(s, str):
        return s
    else:
        encoding = force_encoding(encoding)
        try:
            # In Python 3 str(b'm') returns the string "b'm'" and not just "m",
            # this fixes this.
            return str(s, encoding, "replace")
        except LookupError:
            # The provided enconding is not know, try with no encoding.
            return safe_decode(s)
        except TypeError:
            # For numbers and other stuff.
            return str(s)


def safe_encode(u, encoding=None):
    """Similar to unicode `encode` method returning bytes.

    Encodes `u` using the given `encoding`, or determining one from the system.

    Returning type is always `bytes`; but in python 2.x is also `str`.

    .. versionadded:: 1.1.3
    .. versionchanged:: 1.8.0 migrated to 'future.codecs'

    """
    # XXX: 'eight' pending.
    # TODO: This is not nice for Python 3, bytes is not valid string any more
    #       See `json.encoder.py_encode_basestring_ascii`:func: of Python 2.x
    if isinstance(u, bytes):
        return u
    else:
        encoding = force_encoding(encoding)
        try:
            try:
                if isinstance(u, str):
                    return bytes(u, encoding, "replace")
                else:
                    return str(u).encode(encoding, "replace")
            except (UnicodeError, TypeError):
                return str(u).encode(encoding, "replace")
        except LookupError:
            return safe_encode(u)


__all__ += ("force_encoding", "safe_decode", "safe_encode")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import unittest
from unittest.mock import patch
from datetime import datetime, date

from xotl.tools.records import record, datetime_reader, date_reader

from hypothesis import given
from hypothesis.strategies import composite, text, integers, datetimes


FMT = "%Y-%m-%d"


class _table(record):
    ID = 0
    _id_reader = lambda val: int(val)


class person(_table):
    NAME = 1
    LASTNAME = 2
    BIRTHDATE = 3

    _birthdate_reader = datetime_reader("%Y-%m-%d")

    @property
    def current_age(self):
        from datetime import datetime

        today = datetime.today()
        return self.age_when(today)

    def age_when(self, today):
        res = today - self.birthdate
        return int(res.days // 365.25)


MIN_DATE = datetime(1920, 1, 1)
MAX_DATE = datetime(2007, 12, 31)


@composite
def persons(draw):
    id = draw(integers())
    name = draw(text())
    lastname = draw(text())
    birthday = draw(datetimes(MIN_DATE, MAX_DATE)).strftime(FMT)
    return (
        (id, name, lastname, birthday),
        person((id, name, lastname, birthday)),
    )


class TestRecords(unittest.TestCase):
    def test_records(self):
        from datetime import datetime

        _manu = ("1", "Manuel", "Vazquez", "1978-10-21")
        manu = person(_manu)

        self.assertEqual(1, person.get_field(_manu, person.ID))
        self.assertEqual(1, manu.id)
        self.assertEqual(11, manu.age_when(datetime(1989, 10, 21)))
        self.assertEqual(35, manu.age_when(datetime(2014, 9, 22)))

    @given(p=persons())
    def test_record(self, p):
        r, p = p
        self.assertEqual(r[0], person.get_field(r, person.ID))
        self.assertEqual(r[0], p.id)
        self.assertEqual(r[1], p.name)
        self.assertEqual(r[2], p.lastname)
        self.assertEqual(r[3], p.birthdate.strftime(FMT))

    def test_descriptor(self):
        class INVOICE(record):
            ID = 0
            REFERER = 1

            # The following attribute will be overwritten by the fields
            # descriptor for REFERER.
            referer = "overwritten"

        assert INVOICE.referer and INVOICE.id
        line = (1, "MVA.98")
        self.assertEqual(INVOICE.get_field(line, INVOICE.ID), 1)
        invoice = INVOICE(line)
        self.assertEqual(invoice.referer, "MVA.98")
        self.assertEqual(invoice[INVOICE.REFERER], invoice.referer)

    def test_readers(self):
        from datetime import datetime, timedelta

        class INVOICE(record):
            ID = 0
            REFERER = 1
            CREATED_DATETIME = 2
            UPDATE_DATETIME = 3
            _DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

            @staticmethod
            def _created_datetime_reader(value):
                return datetime.strptime(value, INVOICE._DATETIME_FORMAT)

            # implicit staticmethod
            def _update_datetime_reader(value):
                return datetime.strptime(value, INVOICE._DATETIME_FORMAT)

        today = datetime.now()
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        line = (
            1,
            "MVA.98",
            yesterday.strftime(INVOICE._DATETIME_FORMAT),
            tomorrow.strftime(INVOICE._DATETIME_FORMAT),
        )
        self.assertEqual(INVOICE.get_field(line, INVOICE.CREATED_DATETIME), yesterday)
        self.assertEqual(INVOICE.get_field(line, INVOICE.UPDATE_DATETIME), tomorrow)

        invoice = INVOICE(line)
        self.assertEqual(invoice.created_datetime, yesterday)
        self.assertEqual(invoice.update_datetime, tomorrow)

    def test_default_values(self):
        from xotl.tools.records import float_reader

        class LINE(record):
            DEBIT = "Debit"
            CREDIT = "Credit"
            _debit_reader = float_reader(nullable=True, default=0)
            _credit_reader = float_reader(nullable=True)

        nodata = {}
        partialdata = {"Credit": 0}
        nulls = {"Debit": ""}
        self.assertEqual(LINE.get_field(nodata, LINE.DEBIT), 0)
        self.assertIsNone(LINE.get_field(nodata, LINE.CREDIT))
        self.assertEqual(LINE.get_field(partialdata, LINE.DEBIT), 0)
        self.assertEqual(LINE.get_field(partialdata, LINE.CREDIT), 0)
        self.assertEqual(LINE.get_field(nulls, LINE.DEBIT), 0)


class TestDateTimeReader(unittest.TestCase):
    def setUp(self):
        # clear lru caches for each test... Needed so that imports done inside
        # datetime_reader are mockable.
        datetime_reader.cache_clear()

    def test_strict(self):
        class rec(record):
            MOMENT = 0
            _moment_reader = datetime_reader(FMT)

        inst = rec(["2014-12-17"])
        self.assertEqual("2014-12-17", inst.moment.strftime(FMT))

        inst = rec(["201-12-17"])
        with self.assertRaises(ValueError):
            self.assertEqual("201-12-17", inst.moment.strftime(FMT))

    def test_relaxed_but_nonnullable_with_dateutil(self):
        class rec(record):
            MOMENT = 0
            _moment_reader = datetime_reader(FMT, nullable=False, strict=False)

        inst = rec(["201-12-17"])

        self.assertEqual(inst.moment, datetime(201, 12, 17))

    @patch("dateutil.parser.parse", None)
    def test_relaxed_but_nonnullable_without_dateutil(self):
        class rec(record):
            MOMENT = 0
            _moment_reader = datetime_reader(FMT, nullable=False, strict=False)

        inst = rec(["201-12-17"])
        with self.assertRaises(ValueError):
            self.assertIsNone(inst.moment)

    @patch("dateutil.parser.parse", None)
    def test_relax_with_default(self):
        class rec(record):
            MOMENT = 0
            _moment_reader = datetime_reader(FMT, default=0, strict=False)

        inst = rec(["201-12-17"])
        self.assertEqual(inst.moment, 0)


class TestDateReader(unittest.TestCase):
    def setUp(self):
        # clear lru caches for each test... Needed so that imports done inside
        # date_reader are mockable.
        date_reader.cache_clear()
        datetime_reader.cache_clear()

    def test_date_reader_nullable(self):
        class rec(record):
            WHEN = "date"
            _when_reader = date_reader(FMT, nullable=True)

        inst = rec({"date": "2015-01-01"})
        self.assertEqual(inst.when, date(2015, 1, 1))

        inst = rec({})
        self.assertIsNone(inst.when)

        inst = rec({"date": "201-01-01"})
        with self.assertRaises(ValueError):
            inst.when

    def test_date_reader_relaxed_with_dateutil(self):
        class rec(record):
            WHEN = "date"
            _when_reader = date_reader(FMT, strict=False)

        inst = rec({"date": "201-01-01"})
        self.assertEqual(inst.when, date(201, 1, 1))

    @patch("dateutil.parser.parse", None)
    def test_date_reader_relaxed_nullable_no_dateutil(self):
        class rec(record):
            WHEN = "date"
            _when_reader = date_reader(FMT, nullable=True, strict=False)

        inst = rec({"date": "201-01-01"})
        self.assertIsNone(inst.when)

    @patch("dateutil.parser.parse", None)
    def test_date_reader_relaxed_no_dateutil(self):
        class rec(record):
            WHEN = "date"
            _when_reader = date_reader(FMT, strict=False)

        inst = rec({"date": "201-01-01"})
        with self.assertRaises(ValueError):
            self.assertIsNone(inst.when)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2013-2017 Merchise Autrement [~º/~] and Contributors
# Copyright (c) 2012 Manuel Vazquez

from multiprocessing import Pool, cpu_count
from itertools import count, islice, repeat

try:
    from itertools import izip as zip
except ImportError:
    pass


def gcd(a, b):
    while a % b != 0:
        a, b = b, a % b
    return b


def job(args):
    which, show = args
    try:
        result = gcd(which, which + 2 ** 37 - 73)
        if show and result != 1:
            print(which, result)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    from datetime import datetime
    import argparse

    cpus = max(cpu_count() - 1, 1)
    parser = argparse.ArgumentParser(
        "Run several workers calculating GCD "
        "for large numbers. Uses multiprocessing "
        "to test several CPUs."
    )
    parser.add_argument(
        "--show",
        help="Whether to print or not the gcds",
        action="store_true",
        dest="show",
        default=False,
    )
    parser.add_argument(
        "--workers",
        help="Number of workers. Defaults to %d." % cpus,
        type=int,
        action="store",
        dest="workers",
        default=cpus,
    )
    parser.add_argument(
        "--size",
        help="The magnitude of the test. Represents how many "
        "cycles are processed, exactly 10**MAGNITUDE; "
        "so be nice!",
        type=int,
        action="store",
        dest="magnitude",
        default=6,
    )
    args = parser.parse_args()

    pool = Pool(processes=args.workers)
    start = datetime.now()
    try:
        print("Working...")
        pool.imap(
            job,
            zip(islice(count(2 ** 1028 + 1), 10 ** args.magnitude), repeat(args.show)),
            chunksize=1024,
        )
        pool.close()
        pool.join()
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()
    end = datetime.now()
    total_seconds = (end - start).total_seconds()
    print("All workers (%d) finished in %f seconds" % (args.workers, total_seconds))



from datetime import *
from typing import *

def new_date(d: date) -> date: ...
def new_datetime(d: date) -> datetime: ...
def strfdelta(delta: timedelta) -> str: ...
def strftime(dt: date, fmt: str) -> str: ...
def parse_date(value: str = None) -> date: ...
def parse_datetime(value: str = None) -> datetime: ...
def get_month_first(ref: date = None) -> date: ...
def get_month_last(ref: date = None) -> date: ...
def get_next_month(ref: date = None, lastday: bool = False) -> date: ...
def is_full_month(start: date, end: date) -> bool: ...

class flextime(timedelta):
    @classmethod
    def parse_simple_timeformat(cls, which: str) -> Tuple[int, int, int]: ...

class TimeSpan:
    start_date: Optional[date]
    end_date: Optional[date]
    def __init__(
        self, start_date: Union[str, date] = None, end_date: Union[str, date] = None
    ) -> None: ...
    @classmethod
    def from_date(self, date: date) -> "TimeSpan": ...
    @property
    def past_unbound(self) -> bool: ...
    @property
    def future_unbound(self) -> bool: ...
    @property
    def unbound(self) -> bool: ...
    @property
    def bound(self) -> bool: ...
    @property
    def valid(self) -> bool: ...
    def __contains__(self, other: date) -> bool: ...
    def overlaps(self, other: "TimeSpan") -> bool: ...
    def isdisjoint(self, other: "TimeSpan") -> bool: ...
    def __le__(self, other: "TimeSpan") -> bool: ...
    def issubset(self, other: "TimeSpan") -> bool: ...
    def __lt__(self, other: "TimeSpan") -> bool: ...
    def __gt__(self, other: "TimeSpan") -> bool: ...
    def __ge__(self, other: "TimeSpan") -> bool: ...
    def covers(self, other: "TimeSpan") -> bool: ...
    def issuperset(self, other: "TimeSpan") -> bool: ...
    def __iter__(self) -> Iterator[date]: ...
    def __getitem__(self, index: int) -> date: ...
    def __eq__(self, other) -> bool: ...
    def __and__(self, other: "TimeSpan") -> "TimeSpan": ...
    def __lshift__(self, delta: Union[int, timedelta]) -> "TimeSpan": ...
    def __rshift__(self, delta: Union[int, timedelta]) -> "TimeSpan": ...
    def intersection(self, *others: "TimeSpan") -> "TimeSpan": ...
    def diff(self, other: "TimeSpan") -> Tuple[TimeSpan, TimeSpan]: ...

class DateTimeSpan(TimeSpan):
    start_datetime: Optional[datetime]
    end_datetime: Optional[datetime]
    def __init__(
        self,
        start_datetime: Union[str, date] = None,
        end_datetime: Union[str, date] = None,
    ) -> None: ...
    @classmethod
    def from_date(self, d: date) -> "DateTimeSpan": ...
    @classmethod
    def from_datetime(self, dt: datetime) -> "DateTimeSpan": ...
    @classmethod
    def from_timespan(self, ts: TimeSpan) -> "DateTimeSpan": ...
    def __iter__(self) -> Iterator[datetime]: ...
    def __getitem__(self, index: int) -> datetime: ...
    def __and__(self, other: TimeSpan) -> "DateTimeSpan": ...
    def __lshift__(self, delta: Union[int, timedelta]) -> "DateTimeSpan": ...
    def __rshift__(self, delta: Union[int, timedelta]) -> "DateTimeSpan": ...
    def intersection(self, *others: TimeSpan) -> "DateTimeSpan": ...
    def diff(self, other: TimeSpan) -> Tuple["DateTimeSpan", "DateTimeSpan"]: ...

EmptyTimeSpan: DateTimeSpan



# Copied from Python 3.3 base code
#
# Copyright (c) 2001-2012, 2017 Python Software Foundation.  All rights reserved.
#

from xotl.tools.future import functools
import unittest
from random import choice


class TestLRU(unittest.TestCase):
    def test_lru(self):
        def orig(x, y):
            return 3 * x + y

        f = functools.lru_cache(maxsize=20)(orig)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(maxsize, 20)
        self.assertEqual(currsize, 0)
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 0)

        domain = range(5)
        for i in range(1000):
            x, y = choice(domain), choice(domain)
            actual = f(x, y)
            expected = orig(x, y)
            self.assertEqual(actual, expected)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertTrue(hits > misses)
        self.assertEqual(hits + misses, 1000)
        self.assertEqual(currsize, 20)

        f.cache_clear()  # test clearing
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 0)
        self.assertEqual(currsize, 0)
        f(x, y)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 1)
        self.assertEqual(currsize, 1)

        # Test bypassing the cache
        self.assertIs(f.__wrapped__, orig)
        f.__wrapped__(x, y)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 1)
        self.assertEqual(currsize, 1)

        # test size zero (which means "never-cache")
        @functools.lru_cache(0)
        def f():
            global f_cnt
            f_cnt += 1
            return 20

        self.assertEqual(f.cache_info().maxsize, 0)
        global f_cnt
        f_cnt = 0
        for i in range(5):
            self.assertEqual(f(), 20)
        self.assertEqual(f_cnt, 5)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 0)
        self.assertEqual(misses, 5)
        self.assertEqual(currsize, 0)

        # test size one
        @functools.lru_cache(1)
        def f():
            global f_cnt
            f_cnt += 1
            return 20

        self.assertEqual(f.cache_info().maxsize, 1)
        f_cnt = 0
        for i in range(5):
            self.assertEqual(f(), 20)
        self.assertEqual(f_cnt, 1)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 4)
        self.assertEqual(misses, 1)
        self.assertEqual(currsize, 1)

        # test size two
        @functools.lru_cache(2)
        def f(x):
            global f_cnt
            f_cnt += 1
            return x * 10

        self.assertEqual(f.cache_info().maxsize, 2)
        f_cnt = 0
        for x in 7, 9, 7, 9, 7, 9, 8, 8, 8, 9, 9, 9, 8, 8, 8, 7:
            #    *  *              *                          *
            self.assertEqual(f(x), x * 10)
        self.assertEqual(f_cnt, 4)
        hits, misses, maxsize, currsize = f.cache_info()
        self.assertEqual(hits, 12)
        self.assertEqual(misses, 4)
        self.assertEqual(currsize, 2)

    def test_lru_with_maxsize_none(self):
        @functools.lru_cache(maxsize=None)
        def fib(n):
            if n < 2:
                return n
            return fib(n - 1) + fib(n - 2)

        self.assertEqual(
            [fib(n) for n in range(16)],
            [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610],
        )
        self.assertEqual(
            fib.cache_info(),
            functools._CacheInfo(hits=28, misses=16, maxsize=None, currsize=16),
        )
        fib.cache_clear()
        self.assertEqual(
            fib.cache_info(),
            functools._CacheInfo(hits=0, misses=0, maxsize=None, currsize=0),
        )

    def test_lru_with_exceptions(self):
        # Verify that user_function exceptions get passed through without
        # creating a hard-to-read chained exception.
        # http://bugs.python.org/issue13177
        for maxsize in (None, 100):

            @functools.lru_cache(maxsize)
            def func(i):
                return "abc"[i]

            self.assertEqual(func(0), "a")
            with self.assertRaises(IndexError) as cm:  # noqa
                func(15)
            #  The following is only valid in Py33 PEP 3143
            ##  self.assertIsNone(cm.exception.__context__)

            # Verify that the previous exception did not result in a cached
            # entry
            with self.assertRaises(IndexError):
                func(15)

    def test_lru_with_types(self):
        for maxsize in (None, 100):

            @functools.lru_cache(maxsize=maxsize, typed=True)
            def square(x):
                return x * x

            self.assertEqual(square(3), 9)
            self.assertEqual(type(square(3)), type(9))
            self.assertEqual(square(3.0), 9.0)
            self.assertEqual(type(square(3.0)), type(9.0))
            self.assertEqual(square(x=3), 9)
            self.assertEqual(type(square(x=3)), type(9))
            self.assertEqual(square(x=3.0), 9.0)
            self.assertEqual(type(square(x=3.0)), type(9.0))
            self.assertEqual(square.cache_info().hits, 4)
            self.assertEqual(square.cache_info().misses, 4)



from typing import Any, Dict, ContextManager, TypeVar
from typing_extensions import Protocol

C = TypeVar("C")
K = TypeVar("K")
V = TypeVar("V")

class _ContextData(Dict[K, V]): ...

class _ContextProtocol(Protocol[C, K, V]):
    def __getitem__(self, key: C) -> _ContextData[K, V]: ...
    def __call__(self, key: C) -> ContextManager[_ContextData[K, V]]: ...

context: _ContextProtocol
Context: _ContextProtocol



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import unittest
import pickle
import sys

from xotl.tools.future import types


def test_iscollection():
    # TODO: move this test to equivalent for
    # `xotl.tools.values.simple.logic_collection_coerce`
    from xotl.tools.future.collections import UserList, UserDict

    def is_collection(arg):
        from collections.abc import Iterable, Mapping

        avoid = (Mapping, str)
        return isinstance(arg, Iterable) and not isinstance(arg, avoid)

    assert is_collection("all strings are iterable") is False
    assert is_collection(1) is False
    assert is_collection(range(1)) is True
    assert is_collection({}) is False
    assert is_collection(tuple()) is True
    assert is_collection(set()) is True
    assert is_collection(a for a in range(100)) is True

    class Foobar(UserList):
        pass

    assert is_collection(Foobar()) is True

    class Foobar(UserDict):
        pass

    assert is_collection(Foobar()) is False


class NoneTypeTests(unittest.TestCase):
    "To avoid FlyCheck errors"

    def test_identity(self):
        from xotl.tools.future.types import NoneType

        self.assertIs(NoneType, type(None))


class SimpleNamespaceTests(unittest.TestCase):
    def test_constructor(self):
        ns1 = types.SimpleNamespace()
        ns2 = types.SimpleNamespace(x=1, y=2)
        ns3 = types.SimpleNamespace(**dict(x=1, y=2))

        with self.assertRaises(TypeError):
            types.SimpleNamespace(1, 2, 3)

        self.assertEqual(len(ns1.__dict__), 0)
        self.assertEqual(vars(ns1), {})
        self.assertEqual(len(ns2.__dict__), 2)
        self.assertEqual(vars(ns2), {"y": 2, "x": 1})
        self.assertEqual(len(ns3.__dict__), 2)
        self.assertEqual(vars(ns3), {"y": 2, "x": 1})

    def test_unbound(self):
        ns1 = vars(types.SimpleNamespace())
        ns2 = vars(types.SimpleNamespace(x=1, y=2))

        self.assertEqual(ns1, {})
        self.assertEqual(ns2, {"y": 2, "x": 1})

    def test_underlying_dict(self):
        ns1 = types.SimpleNamespace()
        ns2 = types.SimpleNamespace(x=1, y=2)
        ns3 = types.SimpleNamespace(a=True, b=False)
        mapping = ns3.__dict__
        del ns3

        self.assertEqual(ns1.__dict__, {})
        self.assertEqual(ns2.__dict__, {"y": 2, "x": 1})
        self.assertEqual(mapping, dict(a=True, b=False))

    def test_attrget(self):
        ns = types.SimpleNamespace(x=1, y=2, w=3)

        self.assertEqual(ns.x, 1)
        self.assertEqual(ns.y, 2)
        self.assertEqual(ns.w, 3)
        with self.assertRaises(AttributeError):
            ns.z

    def test_attrset(self):
        ns1 = types.SimpleNamespace()
        ns2 = types.SimpleNamespace(x=1, y=2, w=3)
        ns1.a = "spam"
        ns1.b = "ham"
        ns2.z = 4
        ns2.theta = None

        self.assertEqual(ns1.__dict__, dict(a="spam", b="ham"))
        self.assertEqual(ns2.__dict__, dict(x=1, y=2, w=3, z=4, theta=None))

    def test_attrdel(self):
        ns1 = types.SimpleNamespace()
        ns2 = types.SimpleNamespace(x=1, y=2, w=3)

        with self.assertRaises(AttributeError):
            del ns1.spam
        with self.assertRaises(AttributeError):
            del ns2.spam

        del ns2.y
        self.assertEqual(vars(ns2), dict(w=3, x=1))
        ns2.y = "spam"
        self.assertEqual(vars(ns2), dict(w=3, x=1, y="spam"))
        del ns2.y
        self.assertEqual(vars(ns2), dict(w=3, x=1))

        ns1.spam = 5
        self.assertEqual(vars(ns1), dict(spam=5))
        del ns1.spam
        self.assertEqual(vars(ns1), {})

    @unittest.skipIf(sys.version_info >= (3, 9), "Different order in Python 3.9")
    def test_repr(self):
        ns1 = types.SimpleNamespace(x=1, y=2, w=3)
        ns2 = types.SimpleNamespace()
        ns2.x = str("spam")
        ns2._y = 5
        name = "namespace"

        self.assertEqual(repr(ns1), "{name}(w=3, x=1, y=2)".format(name=name))
        self.assertEqual(repr(ns2), "{name}(_y=5, x='spam')".format(name=name))

    @unittest.skipIf(sys.version_info < (3, 9), "Different order in Python 3.9")
    def test_repr_py39(self):
        ns1 = types.SimpleNamespace(x=1, y=2, w=3)
        ns2 = types.SimpleNamespace()
        ns2.x = str("spam")
        ns2._y = 5
        name = "namespace"

        self.assertEqual(repr(ns1), "{name}(x=1, y=2, w=3)".format(name=name))
        self.assertEqual(repr(ns2), "{name}(x='spam', _y=5)".format(name=name))

    def test_equal(self):
        ns1 = types.SimpleNamespace(x=1)
        ns2 = types.SimpleNamespace()
        ns2.x = 1

        self.assertEqual(types.SimpleNamespace(), types.SimpleNamespace())
        self.assertEqual(ns1, ns2)
        self.assertNotEqual(ns2, types.SimpleNamespace())

    def test_nested(self):
        ns1 = types.SimpleNamespace(a=1, b=2)
        ns2 = types.SimpleNamespace()
        ns3 = types.SimpleNamespace(x=ns1)
        ns2.spam = ns1
        ns2.ham = "?"
        ns2.spam = ns3

        self.assertEqual(vars(ns1), dict(a=1, b=2))
        self.assertEqual(vars(ns2), dict(spam=ns3, ham="?"))
        self.assertEqual(ns2.spam, ns3)
        self.assertEqual(vars(ns3), dict(x=ns1))
        self.assertEqual(ns3.x.a, 1)

    def test_recursive(self):
        ns1 = types.SimpleNamespace(c="cookie")
        ns2 = types.SimpleNamespace()
        ns3 = types.SimpleNamespace(x=1)
        ns1.spam = ns1
        ns2.spam = ns3
        ns3.spam = ns2

        self.assertEqual(ns1.spam, ns1)
        self.assertEqual(ns1.spam.spam, ns1)
        self.assertEqual(ns1.spam.spam, ns1.spam)
        self.assertEqual(ns2.spam, ns3)
        self.assertEqual(ns3.spam, ns2)
        self.assertEqual(ns2.spam.spam, ns2)

    @unittest.skipIf(sys.version_info >= (3, 9), "Different order in Python 3.9")
    def test_recursive_repr(self):
        ns1 = types.SimpleNamespace(c=str("cookie"))
        ns2 = types.SimpleNamespace()
        ns3 = types.SimpleNamespace(x=1)
        ns1.spam = ns1
        ns2.spam = ns3
        ns3.spam = ns2
        name = "namespace"
        repr1 = "{name}(c='cookie', spam={name}(...))".format(name=name)
        repr2 = "{name}(spam={name}(spam={name}(...), x=1))".format(name=name)

        self.assertEqual(repr(ns1), repr1)
        self.assertEqual(repr(ns2), repr2)

    @unittest.skipIf(sys.version_info < (3, 9), "Different order in Python 3.9")
    def test_recursive_repr_py39(self):
        ns1 = types.SimpleNamespace(c=str("cookie"))
        ns2 = types.SimpleNamespace()
        ns3 = types.SimpleNamespace(x=1)
        ns1.spam = ns1
        ns2.spam = ns3
        ns3.spam = ns2
        name = "namespace"
        repr1 = "{name}(c='cookie', spam={name}(...))".format(name=name)
        repr2 = "{name}(spam={name}(x=1, spam={name}(...)))".format(name=name)

        self.assertEqual(repr(ns1), repr1)
        self.assertEqual(repr(ns2), repr2)

    def test_as_dict(self):
        ns = types.SimpleNamespace(spam="spamspamspam")

        with self.assertRaises(TypeError):
            len(ns)
        with self.assertRaises(TypeError):
            iter(ns)
        with self.assertRaises(TypeError):
            "spam" in ns
        with self.assertRaises(TypeError):
            ns["spam"]

    def test_subclass(self):
        class Spam(types.SimpleNamespace):
            pass

        spam = Spam(ham=8, eggs=9)

        self.assertIs(type(spam), Spam)
        self.assertEqual(vars(spam), {"ham": 8, "eggs": 9})

    def test_pickle(self):
        ns = types.SimpleNamespace(breakfast="spam", lunch="spam")

        for protocol in range(pickle.HIGHEST_PROTOCOL + 1):
            pname = "protocol {}".format(protocol)
            try:
                ns_pickled = pickle.dumps(ns, protocol)
            except TypeError:
                raise TypeError(pname)
            ns_roundtrip = pickle.loads(ns_pickled)

            self.assertEqual(ns, ns_roundtrip, pname)


class TestDynamicClassAttribute(unittest.TestCase):
    def test_isimportable(self):
        from xotl.tools.future.types import DynamicClassAttribute  # noqa



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import pytest
from xotl.tools.tasking import retry


class FailingMock:
    def __init__(self, start=0, threshold=5, sleeper=0):
        self.start = start
        self.threshold = threshold
        self.sleeper = sleeper

    def __call__(self, incr=1):
        if self.sleeper:
            import time

            time.sleep(self.sleeper)
        self.start += incr
        if self.start < self.threshold:
            raise ValueError(self.start)
        else:
            return self.start


def test_retrying_max_tries_not_enough_tries():
    fn = FailingMock(threshold=1000000, sleeper=0.5)
    with pytest.raises(ValueError):
        retry(fn, max_tries=3)


def test_retrying_max_tries_enough_tries():
    fn = FailingMock()
    assert retry(fn, max_tries=10) == 5


def test_retrying_max_time_runout():
    fn = FailingMock(threshold=1000000, sleeper=0.2)
    with pytest.raises(ValueError):
        retry(fn, max_time=0.3)


def test_retrying_max_time():
    fn = FailingMock(sleeper=0.2)
    assert retry(fn, max_time=0.2 * 5) == 5


def test_signature():
    fn = FailingMock()
    with pytest.raises(TypeError):
        assert retry(fn, (), {}, 5)
    with pytest.raises(TypeError):
        assert retry(fn, (), {}, x=5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Enhanced data pretty printer."""

from pprint import *  # noqa
from pprint import __all__, pprint  # noqa

__all__ = list(__all__) + ["ppformat"]


def ppformat(obj):
    """Just like `pprint`:func: but always returning a result.

    :returns: The pretty formated text.
    :rtype: `unicode` in Python 2, `str` in Python 3.

    """
    import io

    stream = io.StringIO()
    pprint(obj, stream=stream)
    stream.seek(0)
    res = stream.read()
    if isinstance(res, str):
        return res
    else:
        from xotl.tools.future.codecs import safe_decode

        return safe_decode(res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Simple or internal coercers.

With coercers defined in this module, many of the `xotl.tools.string`:mod:
utilities could be deprecated.

In Python 3, all arrays, not only those containing valid byte or unicode
chars, are buffers.

"""

from xotl.tools.values import coercer, nil


@coercer
def not_false_coercer(arg):
    """Validate that `arg` is not a false value.

    Python convention for values considered True or False is not used here,
    our false values are only `None` or any false instance of
    `xotl.tools.symbols.boolean`:class: (of course including `False` itself).

    """
    from xotl.tools.symbols import boolean

    false = arg is None or (not arg and isinstance(arg, boolean))
    return arg if not false else nil


def not_false(default):
    """Create a coercer that returns `default` if `arg` is considered false.

    See `not_false_coercer`:func: for more information on values considered
    false.

    """

    @coercer
    def inner_coercer(arg):
        coercer = not_false_coercer
        return arg if coercer(arg) is arg else coercer(default)

    return inner_coercer


def isnot(value):
    """Create a coercer that returns `arg` if `arg` is not `value`."""

    @coercer
    def inner_coercer(arg):
        return arg if arg is not value else nil

    return inner_coercer


@coercer
def name_coerce(arg):
    """If `arg` is a named object, return its name, else `nil`.

    Object names are always of `str` type, other types are considered
    invalid.

    Generator objects has the special `__name__` attribute, but they are
    ignored and considered invalid.

    """
    from types import GeneratorType

    if isinstance(arg, GeneratorType):
        return nil
    else:
        if isinstance(arg, (staticmethod, classmethod)):
            fn = getattr(arg, "__func__", None)
            if fn:
                arg = fn
        res = getattr(arg, "__name__", None)
        return res if isinstance(res, str) else nil


@coercer
def iterable_coerce(arg):
    """Return the same argument if it is an iterable."""
    from collections.abc import Iterable

    return arg if isinstance(arg, Iterable) else nil


def collection(arg=nil, avoid=(), force=False, base=None, name=None):
    """Coercer for logic collections.

    Inner coercer returns the same argument if it is a strict iterable.  In
    Python, strings are normally iterables, but never in our logic.  So::

      >>> collection('abc') is nil
      True

    This function could directly check an argument if it isn't ``nil``, or
    returns a coercer using extra parameters:

    :param avoid: a type or tuple of extra types to ignore as valid
           collections; for example::

             >>> collection(avoid=dict)({}) is nil
             True
             >>> collection()({}) is nil
             False

    :param force: if main argument is not a valid collection, it is are
           wrapped inner a list::

             >>> collection(avoid=(dict,), force=True)({}) == [{}]
             True

    :param base: if not ``None``, must be the base to check instead of
           `~collections.abc.Iterable`:class:.

    :param name: decorate inner coercer with that function name.

    """
    if not base:
        from collections.abc import Iterable as base
    if not isinstance(avoid, tuple):
        avoid = (avoid,)

    @coercer
    def collection_coerce(arg):
        invalid = (str,) + avoid
        ok = not isinstance(arg, invalid) and isinstance(arg, base)
        return arg if ok else ([arg] if force else nil)

    if arg is nil:
        doc = (
            "Return the same argument if it is a strict iterable.\n    "
            "Strings{} are not considered valid iterables in this case.\n"
        ).format(" and {}".format(avoid) if avoid else "")
        if force:
            doc += "    A non iterable argument is wrapped in a list.\n"
        collection_coerce.__doc__ = doc
        del doc
        if name:
            collection_coerce.__name__ = name
        return collection_coerce
    else:
        assert not name
        return collection_coerce(arg)


from collections.abc import Mapping, Sequence  # noqa

logic_iterable_coerce = collection(name="logic_iterable_coerce")
force_iterable_coerce = collection(force=True, name="force_iterable_coerce")
logic_collection_coerce = collection(avoid=Mapping, name="logic_collection_coerce")
force_collection_coerce = collection(
    avoid=Mapping, force=True, name="force_collection_coerce"
)
logic_sequence_coerce = collection(
    avoid=Mapping, base=Sequence, name="logic_sequence_coerce"
)
force_sequence_coerce = collection(
    avoid=Mapping, force=True, base=Sequence, name="force_sequence_coerce"
)
del Mapping, Sequence


@coercer
def decode_coerce(arg):
    """Decode objects implementing the buffer protocol."""
    import locale

    encoding = locale.getpreferredencoding() or "UTF-8"
    decode = getattr(arg, "decode", None)
    if callable(decode):
        try:
            res = decode(encoding, "replace")
            if not isinstance(res, str):
                res = None
        except Exception:
            res = None
    else:
        res = None
    if res is None:
        try:
            # TODO: All arrays are decoded, and not only those containing
            # valid byte or unicode characters.
            import codecs

            res = codecs.decode(arg, encoding, "replace")
        except Exception:
            res = nil
    return res


@coercer
def encode_coerce(arg):
    """Encode string objects."""
    import locale

    encoding = locale.getpreferredencoding() or "UTF-8"
    encode = getattr(arg, "encode", None)
    if callable(encode):
        try:
            res = encode(encoding, "replace")
            if not isinstance(res, bytes):
                res = None
        except Exception:
            res = None
    else:
        res = None
    if res is None:
        try:
            import codecs

            res = codecs.encode(arg, encoding, "replace")
        except Exception:
            res = nil
    return res


@coercer
def unicode_coerce(arg):
    """Decode a buffer or any object returning unicode text.

    Uses the defined `encoding` system value.

    In Python 2.x unicode has a special type different to `str` but in Python
    3 coincide with `str` type.

    Name is used in named objects, see `name_coerce`:func: for more
    information.

    See `str_coerce`:func: to coerce to standard string type, `bytes` in
    Python 2.x and unicode (`str`) in Python 3.

    .. versionadded:: 1.7.0

    """
    from array import array

    aux = name_coerce(arg)
    if aux is not nil:
        arg = aux
    if isinstance(arg, str):
        return arg
    elif isinstance(arg, bytearray):
        arg = bytes(arg)
    elif isinstance(arg, memoryview):
        arg = arg.tobytes()
    elif isinstance(arg, array):
        try:
            return arg.tounicode()
        except Exception:
            try:
                arg = bytes(bytearray(arg.tolist()))
            except Exception:
                arg = str(arg)
            return arg

    res = decode_coerce(arg)
    return str(arg) if res is nil else res


@coercer
def bytes_coerce(arg):
    """Encode an unicode string (or any object) returning a bytes buffer.

    Uses the defined `encoding` system value.

    In Python 2.x `bytes` coincide with `str` type, in Python 3 `str` uses
    unicode and `str` is different to `bytes`.

    There are differences if you want to obtain a buffer in Python 2.x and
    Python 3; for example, the following code obtain different results::

      >>> ba = bytes([65, 66, 67])

    In Python 2.x is obtained the string ``"[65, 66, 67]"`` and in Python 3
    ``b"ABC"``.  This function normalize these differences.

    Name is used in named objects, see `name_coerce`:func: for more
    information.

    See `str_coerce`:func: to coerce to standard string type, `bytes` in
    Python 2.x and unicode (`str`) in Python 3.

    Always returns the `bytes` type.

    .. versionadded:: 1.7.0

    """
    from array import array

    aux = name_coerce(arg)
    if aux is not nil:
        arg = aux
    if isinstance(arg, bytes):
        return arg
    elif isinstance(arg, bytearray):
        return bytes(arg)
    elif isinstance(arg, memoryview):
        return arg.tobytes()
    elif isinstance(arg, array):
        try:
            arg = arg.tounicode()
        except Exception:
            try:
                return bytes(bytearray(arg.tolist()))
            except Exception:
                arg = str(arg)
    res = encode_coerce(arg)
    return encode_coerce(str(arg)) if res is nil else res


@coercer
def str_coerce(arg):
    """Coerce to standard string type.

    `bytes` in Python 2.x and unicode (`str`) in Python 3.

    .. versionadded:: 1.7.0

    .. deprecated:: 2.0.6

    """
    return unicode_coerce(arg)


@coercer
def ascii_coerce(arg):
    """Coerce to string containing only ASCII characters.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    """
    import unicodedata

    if not isinstance(arg, str):
        arg = unicode_coerce(arg)
    res = unicodedata.normalize("NFKD", arg).encode("ascii", "ignore")
    return str_coerce(res)


@coercer
def ascii_set_coerce(arg):
    """Coerce to string with only ASCII characters removing repetitions.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    """
    return "".join(set(ascii_coerce(arg)))


@coercer
def lower_ascii_coerce(arg):
    """Coerce to string containing only lower-case ASCII characters.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    """
    return ascii_coerce(arg).lower()


@coercer
def lower_ascii_set_coerce(arg):
    """Coerce to string with only lower-case ASCII chars removing repetitions.

    Convert all non-ascii to valid characters using unicode 'NFKC'
    normalization.

    """
    return "".join(set(lower_ascii_coerce(arg)))


@coercer
def chars_coerce(arg):
    """Convert to unicode characters.

    If `arg` is an integer between ``0`` and ``0x10ffff`` is converted
    assuming it as ordinal unicode code, else is converted with
    `unicode_coerce`:meth:.

    """
    if isinstance(arg, int) and 0 <= arg <= 0x10FFFF:
        return chr(arg)
    else:
        return unicode_coerce(arg)


@coercer
def strict_string_coerce(arg):
    """Coerce to string only if argument is a valid string type."""
    return str_coerce(arg) if isinstance(arg, str) else nil


# TODO: Why is this here
class text(str):
    """Return a nice text representation of one object.

    text(obj='') -> text

    text(bytes_or_buffer[, encoding[, errors]]) -> text

    Create a new string object from the given object.  If `encoding` or
    `errors` is specified, then the object must expose a data buffer that will
    be decoded using the given encoding and error handler.  Otherwise, returns
    the result of object text representation.

    :param encoding: defaults to ``sys.getdefaultencoding()``.

    :param errors: defaults to 'strict'.

    Method join is improved, in order to receive any collection of objects,
    as variable number of arguments or as one iterable.

    """

    def __new__(cls, obj="", *args, **kwargs):
        if not (args or kwargs):
            obj = unicode_coerce(obj)
        return super().__new__(cls, obj, *args, **kwargs)

    def join(self, *args):
        """S.join(variable_number_args or iterable) -> text

        Return a text which is the concatenation of the objects (converted to
        text) in argument items.  The separator between elements is `S`.

        See `chr_join`:meth: for other vertion of this functionality.

        """
        return self._join(unicode_coerce, args)

    def chr_join(self, *args):
        """S.chr_join(variable_number_args or iterable) -> text

        Return a text which is the concatenation of the objects (converted to
        text) in argument items.  The separator between elements is `S`.

        Difference with `join`:meth: is that integers between ``0`` and
        ``0x10ffff`` are converted to characters as unicode ordinal.

        """
        return self._join(chars_coerce, args)

    def _join(self, coercer, args):
        """Protected method to implement `join`:meth: and `chr_join`:meth:."""
        from collections.abc import Iterable

        if len(args) == 1 and isinstance(args[0], Iterable):
            args = args[0]
        return super().join(coercer(obj) for obj in args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Backport of Python 3.9's grahlib module.

.. versionadded:: 2.1.9

"""

try:
    from graphlib import *  # noqa
except ImportError:
    from ._graphlib import *  # noqa
